# Copyright 2020 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import os
import re

from oslo_log import log as logging

from backup import utils
from backup.drivers import base
from backup.utils import postgresql as psql_util

LOG = logging.getLogger(__name__)


class PgBasebackup(base.BaseRunner):
    def __init__(self, *args, **kwargs):
        if not kwargs.get('wal_archive_dir'):
            raise AttributeError('wal_archive_dir attribute missing')
        self.wal_archive_dir = kwargs.pop('wal_archive_dir')
        self.datadir = kwargs.pop(
            'db_datadir', '/var/lib/postgresql/data/pgdata')

        self.label = None
        self.stop_segment = None
        self.start_segment = None
        self.start_wal_file = None
        self.stop_wal_file = None
        self.checkpoint_location = None
        self.metadata = {}

        super(PgBasebackup, self).__init__(*args, **kwargs)

        self.restore_command = (f"{self.decrypt_cmd}tar xzf - -C "
                                f"{self.datadir}")

    @property
    def cmd(self):
        cmd = (f"pg_basebackup -U postgres -Ft -z --wal-method=fetch "
               f"--label={self.filename} --pgdata=-")
        return cmd + self.encrypt_cmd

    @property
    def manifest(self):
        """Target file name."""
        return "%s.tar.gz%s" % (self.filename, self.encrypt_manifest)

    def get_wal_files(self, backup_pos=0):
        """Return the WAL files since the provided last backup.

        pg_archivebackup depends on alphanumeric sorting to decide wal order,
        so we'll do so too:
        https://github.com/postgres/postgres/blob/REL9_4_STABLE/contrib
           /pg_archivecleanup/pg_archivecleanup.c#L122
        """
        backup_file = self.get_backup_file(backup_pos=backup_pos)
        last_wal = backup_file.split('.')[0]
        wal_re = re.compile("^[0-9A-F]{24}$")
        wal_files = [wal_file for wal_file in os.listdir(self.wal_archive_dir)
                     if wal_re.search(wal_file) and wal_file >= last_wal]
        return wal_files

    def get_backup_file(self, backup_pos=0, regex=None):
        """Look for the most recent .backup file that basebackup creates

        :return: a string like 000000010000000000000006.00000168.backup
        """
        regex = regex or r"[0-9A-F]{24}\..*\.backup"
        backup_re = re.compile(regex)
        wal_files = [wal_file for wal_file in os.listdir(self.wal_archive_dir)
                     if backup_re.search(wal_file)]
        wal_files = sorted(wal_files, reverse=True)
        if not wal_files:
            return None
        return wal_files[backup_pos]

    def get_backup_metadata(self, metadata_file):
        """Parse the contents of the .backup file"""
        metadata = {}

        start_re = re.compile(r"START WAL LOCATION: (.*) \(file (.*)\)")
        stop_re = re.compile(r"STOP WAL LOCATION: (.*) \(file (.*)\)")
        checkpt_re = re.compile("CHECKPOINT LOCATION: (.*)")
        label_re = re.compile("LABEL: (.*)")

        with open(metadata_file, 'r') as file:
            metadata_contents = file.read()

        match = start_re.search(metadata_contents)
        if match:
            self.start_segment = match.group(1)
            metadata['start-segment'] = self.start_segment
            self.start_wal_file = match.group(2)
            metadata['start-wal-file'] = self.start_wal_file

        match = stop_re.search(metadata_contents)
        if match:
            self.stop_segment = match.group(1)
            metadata['stop-segment'] = self.stop_segment
            self.stop_wal_file = match.group(2)
            metadata['stop-wal-file'] = self.stop_wal_file

        match = checkpt_re.search(metadata_contents)
        if match:
            self.checkpoint_location = match.group(1)
            metadata['checkpoint-location'] = self.checkpoint_location

        match = label_re.search(metadata_contents)
        if match:
            self.label = match.group(1)
            metadata['label'] = self.label

        return metadata

    def get_metadata(self):
        """Get metadata.

        pg_basebackup may complete, and we arrive here before the
        history file is written to the wal archive. So we need to
        handle two possibilities:
        - this is the first backup, and no history file exists yet
        - this isn't the first backup, and so the history file we retrieve
        isn't the one we just ran!
         """
        def _metadata_found():
            backup_file = self.get_backup_file()
            if not backup_file:
                return False

            self.metadata = self.get_backup_metadata(
                os.path.join(self.wal_archive_dir, backup_file))
            LOG.info("Metadata for backup: %s.", self.metadata)
            return self.metadata['label'] == self.filename

        try:
            LOG.debug("Polling for backup metadata... ")
            utils.poll_until(_metadata_found, sleep_time=5, time_out=60)
        except Exception as e:
            raise RuntimeError(f"Failed to get backup metadata for backup "
                               f"{self.filename}: {str(e)}")

        return self.metadata

    def check_process(self):
        # If any of the below variables were not set by either metadata()
        # or direct retrieval from the pgsql backup commands, then something
        # has gone wrong
        if not self.start_segment or not self.start_wal_file:
            LOG.error("Unable to determine starting WAL file/segment")
            return False
        if not self.stop_segment or not self.stop_wal_file:
            LOG.error("Unable to determine ending WAL file/segment")
            return False
        if not self.label:
            LOG.error("No backup label found")
            return False
        return True


class PgBasebackupIncremental(PgBasebackup):
    """Incremental backup/restore for PostgreSQL.

    To restore an incremental backup from a previous backup, in PostgreSQL,
    is effectively to replay the WAL entries to a designated point in time.
    All that is required is the most recent base backup, and all WAL files
    """

    def __init__(self, *args, **kwargs):
        self.parent_location = kwargs.pop('parent_location', '')
        self.parent_checksum = kwargs.pop('parent_checksum', '')
        self.parent_stop_wal = kwargs.pop('stop_wal_file', '')

        super(PgBasebackupIncremental, self).__init__(*args, **kwargs)

        self.incr_restore_cmd = f'tar -xzf - -C {self.wal_archive_dir}'

    def pre_backup(self):
        # Check if the parent stop wal file still exists. It may be removed
        # by trove-guestagent.
        parent_wal_name = self.get_backup_file(
            backup_pos=0, regex=fr'{self.parent_stop_wal}\..+\.backup')
        if not parent_wal_name:
            raise Exception("Cannot find parent backup WAL file.")

        with psql_util.PostgresConnection('postgres') as conn:
            self.start_segment = conn.query(
                f"SELECT pg_start_backup('{self.filename}', false, false)"
            )[0][0]
            self.start_wal_file = conn.query(
                f"SELECT pg_walfile_name('{self.start_segment}')")[0][0]
            self.stop_segment = conn.query(
                "SELECT * FROM pg_stop_backup(false, true)")[0][0]

        # We have to hack this because self.command is
        # initialized in the base class before we get here, which is
        # when we will know exactly what WAL files we want to archive
        self.command = self._cmd()

    def _cmd(self):
        wal_file_list = self.get_wal_files(backup_pos=1)
        cmd = (f'tar -czf - -C {self.wal_archive_dir} '
               f'{" ".join(wal_file_list)}')
        return cmd + self.encrypt_cmd

    def get_metadata(self):
        _meta = super(PgBasebackupIncremental, self).get_metadata()
        _meta.update({
            'parent_location': self.parent_location,
            'parent_checksum': self.parent_checksum,
        })
        return _meta

    def incremental_restore_cmd(self, incr=False):
        cmd = self.restore_command
        if incr:
            cmd = self.incr_restore_cmd
        return self.decrypt_cmd + cmd

    def incremental_restore(self, location, checksum):
        """Perform incremental restore.

        For the child backups, restore the wal files to wal archive dir.
        For the base backup, restore to datadir.
        """
        metadata = self.storage.load_metadata(location, checksum)
        if 'parent_location' in metadata:
            LOG.info("Restoring parent: %(parent_location)s, "
                     "checksum: %(parent_checksum)s.", metadata)

            parent_location = metadata['parent_location']
            parent_checksum = metadata['parent_checksum']

            # Restore parents recursively so backup are applied sequentially
            self.incremental_restore(parent_location, parent_checksum)

            command = self.incremental_restore_cmd(incr=True)
        else:
            # For the parent base backup, revert to the default restore cmd
            LOG.info("Restoring back to full backup.")
            command = self.incremental_restore_cmd(incr=False)

        self.restore_content_length += self.unpack(location, checksum, command)

    def run_restore(self):
        """Run incremental restore."""
        LOG.debug('Running incremental restore')
        self.incremental_restore(self.location, self.checksum)
        return self.restore_content_length



# Copyright [2015] Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from trove.common.clients import create_guest_client
from trove.common.db import models as guest_models
from trove.common import exception
from trove.common import timeutils
from trove.db import get_db_api
from trove.instance import models as base_models


LOG = logging.getLogger(__name__)


def load_and_verify(context, instance_id,
                    enabled_datastore=['mysql', 'mariadb']):
    """Check instance datastore.

    Some API operations are only supported for some specific datastores.
    """
    instance = base_models.Instance.load(context, instance_id)

    if instance.datastore_version.manager not in enabled_datastore:
        raise exception.UnprocessableEntity(
            "Operation not supported for datastore "
            f"{instance.datastore_version.manager}."
        )

    if not instance.is_datastore_running:
        raise exception.UnprocessableEntity(
            "Instance %s is not ready, status: %s." %
            (instance.id, instance.datastore_status.status)
        )

    return instance


class Root(object):

    @classmethod
    def load(cls, context, instance_id):
        load_and_verify(context, instance_id,
                        enabled_datastore=['mysql', 'mariadb', 'postgresql'])
        # TODO(pdmars): remove the is_root_enabled call from the guest agent,
        # just check the database for this information.
        # If the root history returns null or raises an exception, the root
        # user hasn't been enabled.
        try:
            root_history = RootHistory.load(context, instance_id)
        except exception.NotFound:
            return False
        if not root_history:
            return False
        return True

    @classmethod
    def create(cls, context, instance_id, root_password,
               cluster_instances_list=None):
        load_and_verify(context, instance_id,
                        enabled_datastore=['mysql', 'mariadb', 'postgresql'])
        if root_password:
            root = create_guest_client(context,
                                       instance_id).enable_root_with_password(
                root_password)
        else:
            root = create_guest_client(context, instance_id).enable_root()

        root_user = guest_models.DatastoreUser.deserialize(root,
                                                           verify=False)
        root_user.make_root()

        # if cluster_instances_list none, then root create is called for
        # single instance, adding an RootHistory entry for the instance_id
        if cluster_instances_list is None:
            RootHistory.create(context, instance_id)

        return root_user

    @classmethod
    def delete(cls, context, instance_id):
        load_and_verify(context, instance_id,
                        enabled_datastore=['mysql', 'mariadb', 'postgresql'])
        create_guest_client(context, instance_id).disable_root()


class ClusterRoot(Root):

    @classmethod
    def create(cls, context, instance_id, root_password,
               cluster_instances_list=None):
        root_user = super(ClusterRoot, cls).create(context, instance_id,
                                                   root_password,
                                                   cluster_instances_list=None)

        if cluster_instances_list:
            for instance in cluster_instances_list:
                RootHistory.create(context, instance)

        return root_user


class RootHistory(object):

    _auto_generated_attrs = ['id']
    _data_fields = ['instance_id', 'user', 'created']
    _table_name = 'root_enabled_history'

    def __init__(self, instance_id, user):
        self.id = instance_id
        self.user = user
        self.created = timeutils.utcnow()

    def save(self):
        LOG.debug("Saving %(name)s: %(dict)s",
                  {'name': self.__class__.__name__, 'dict': self.__dict__})
        return get_db_api().save(self)

    @classmethod
    def load(cls, context, instance_id):
        history = get_db_api().find_by(cls, id=instance_id)
        return history

    @classmethod
    def create(cls, context, instance_id):
        history = cls.load(context, instance_id)
        if history is not None:
            return history
        history = RootHistory(instance_id, context.user_id)
        return history.save()



# Copyright [2015] Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class DatastoreVersionView(object):

    def __init__(self, datastore_version):
        self.datastore_version = datastore_version

    def data(self):
        datastore_version_dict = {
            "id": self.datastore_version.id,
            "name": self.datastore_version.name,
            "version": self.datastore_version.version,
            "datastore_id": self.datastore_version.datastore_id,
            "datastore_name": self.datastore_version.datastore_name,
            "datastore_manager": self.datastore_version.manager,
            "image": self.datastore_version.image_id,
            "image_tags": (self.datastore_version.image_tags.split(',')
                           if self.datastore_version.image_tags else ['']),
            "packages": (self.datastore_version.packages.split(
                ',') if self.datastore_version.packages else ['']),
            "active": self.datastore_version.active,
            "default": self.datastore_version.default}

        return {'version': datastore_version_dict}


class DatastoreVersionsView(object):

    def __init__(self, datastore_versions):
        self.datastore_versions = datastore_versions

    def data(self):
        data = []
        for datastore_version in self.datastore_versions:
            data.append(
                DatastoreVersionView(datastore_version).data()['version'])

        return {'versions': data}



# Copyright (c) 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
:mod:`tests` -- Utility methods for tests.
===================================

.. automodule:: utils
   :platform: Unix
   :synopsis: Tests for Nova.
"""

import subprocess
from urllib.parse import unquote
try:
    EVENT_AVAILABLE = True
except ImportError:
    EVENT_AVAILABLE = False

import glanceclient
from keystoneauth1.identity import v3
from keystoneauth1 import session
from neutronclient.v2_0 import client as neutron_client
from novaclient import client as nova_client
from proboscis.asserts import assert_true
from proboscis.asserts import Check
from proboscis.asserts import fail
from proboscis import SkipTest
from sqlalchemy import create_engine
from sqlalchemy.sql.expression import text
import tenacity
from troveclient.compat import Dbaas

from trove.common import cfg
from trove.common.utils import import_class
from trove.common.utils import import_object
from trove.tests.config import CONFIG as test_config
from trove.tests.util.client import TestClient
from trove.tests.util import mysql
from trove.tests.util import test_config as CONFIG
from trove.tests.util.users import Requirements


WHITE_BOX = test_config.white_box
FLUSH = text("FLUSH PRIVILEGES;")
CONF = cfg.CONF


def create_client(*args, **kwargs):
    """
    Using the User Requirements as arguments, finds a user and grabs a new
    DBAAS client.
    """
    reqs = Requirements(*args, **kwargs)
    user = test_config.users.find_user(reqs)
    return create_dbaas_client(user)


def create_dbaas_client(user):
    """Creates a rich client for the Trove API using the test config."""
    auth_strategy = None

    kwargs = {
        'service_type': 'database',
        'insecure': test_config.values['trove_client_insecure'],
    }

    def set_optional(kwargs_name, test_conf_name):
        value = test_config.values.get(test_conf_name, None)
        if value is not None:
            kwargs[kwargs_name] = value
    force_url = 'override_trove_api_url' in test_config.values

    service_url = test_config.get('override_trove_api_url', None)
    if user.requirements.is_admin:
        service_url = test_config.get('override_admin_trove_api_url',
                                      service_url)
    if service_url:
        kwargs['service_url'] = service_url

    auth_strategy = None
    if user.requirements.is_admin:
        auth_strategy = test_config.get('admin_auth_strategy',
                                        test_config.auth_strategy)
    else:
        auth_strategy = test_config.auth_strategy
    set_optional('region_name', 'trove_client_region_name')
    if test_config.values.get('override_trove_api_url_append_tenant',
                              False):
        kwargs['service_url'] += "/" + user.tenant

    if auth_strategy == 'fake':
        from troveclient.compat import auth

        class FakeAuth(auth.Authenticator):
            def authenticate(self):
                class FakeCatalog(object):
                    def __init__(self, auth):
                        self.auth = auth

                    def get_public_url(self):
                        return "%s/%s" % (test_config.dbaas_url,
                                          self.auth.tenant)

                    def get_token(self):
                        return self.auth.tenant

                return FakeCatalog(self)

        auth_strategy = FakeAuth

    if auth_strategy:
        kwargs['auth_strategy'] = auth_strategy

    if not user.requirements.is_admin:
        auth_url = test_config.trove_auth_url
    else:
        auth_url = test_config.values.get('trove_admin_auth_url',
                                          test_config.trove_auth_url)

    if test_config.values.get('trove_client_cls'):
        cls_name = test_config.trove_client_cls
        kwargs['client_cls'] = import_class(cls_name)

    dbaas = Dbaas(user.auth_user, user.auth_key, tenant=user.tenant,
                  auth_url=auth_url, **kwargs)
    dbaas.authenticate()
    with Check() as check:
        check.is_not_none(dbaas.client.auth_token, "Auth token not set!")
        if not force_url and user.requirements.is_admin:
            expected_prefix = test_config.dbaas_url
            actual = dbaas.client.service_url
            msg = "Dbaas management url was expected to start with %s, but " \
                  "was %s." % (expected_prefix, actual)
            check.true(actual.startswith(expected_prefix), msg)
    return TestClient(dbaas)


def create_keystone_session(user):
    auth = v3.Password(username=user.auth_user,
                       password=user.auth_key,
                       project_id=user.tenant_id,
                       user_domain_name='Default',
                       project_domain_name='Default',
                       auth_url=test_config.auth_url)
    return session.Session(auth=auth)


def create_nova_client(user, service_type=None):
    if not service_type:
        service_type = CONF.nova_compute_service_type
    openstack = nova_client.Client(
        CONF.nova_client_version,
        username=user.auth_user,
        password=user.auth_key,
        user_domain_name='Default',
        project_id=user.tenant_id,
        auth_url=CONFIG.auth_url,
        service_type=service_type, os_cache=False,
        cacert=test_config.values.get('cacert', None)
    )

    return TestClient(openstack)


def create_neutron_client(user):
    sess = create_keystone_session(user)
    client = neutron_client.Client(
        session=sess,
        service_type=CONF.neutron_service_type,
        region_name=CONFIG.trove_client_region_name,
        insecure=CONF.neutron_api_insecure,
        endpoint_type=CONF.neutron_endpoint_type
    )

    return TestClient(client)


def create_glance_client(user):
    sess = create_keystone_session(user)
    glance = glanceclient.Client(CONF.glance_client_version, session=sess)

    return TestClient(glance)


def dns_checker(mgmt_instance):
    """Given a MGMT instance, ensures DNS provisioning worked.

    Uses a helper class which, given a mgmt instance (returned by the mgmt
    API) can confirm that the DNS record provisioned correctly.
    """
    if CONFIG.values.get('trove_dns_checker') is not None:
        checker = import_class(CONFIG.trove_dns_checker)
        checker()(mgmt_instance)
    else:
        raise SkipTest("Can't access DNS system to check if DNS provisioned.")


def process(cmd):
    output = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
    return output


def string_in_list(str, substr_list):
    """Returns True if the string appears in the list."""
    return any([str.find(x) >= 0 for x in substr_list])


def unquote_user_host(user_hostname):
    unquoted = unquote(user_hostname)
    if '@' not in unquoted:
        return unquoted, '%'
    if unquoted.endswith('@'):
        return unquoted, '%'
    splitup = unquoted.split('@')
    host = splitup[-1]
    user = '@'.join(splitup[:-1])
    return user, host


def iso_time(time_string):
    """Return a iso formated datetime: 2013-04-15T19:50:23Z."""
    ts = time_string.replace(' ', 'T')
    try:
        micro = ts.rindex('.')
        ts = ts[:micro]
    except ValueError:
        pass
    return '%sZ' % ts


def assert_contains(exception_message, substrings):
    for substring in substrings:
        assert_true(substring in exception_message,
                    message="'%s' not in '%s'"
                    % (substring, exception_message))


# TODO(dukhlov): Still required by trove integration
# Should be removed after trove integration fix
# https://bugs.launchpad.net/trove-integration/+bug/1228306


# TODO(cp16net): DO NOT USE needs to be removed
def mysql_connection():
    cls = CONFIG.get('mysql_connection',
                     "local.MySqlConnection")
    if cls == "local.MySqlConnection":
        return MySqlConnection()
    return import_object(cls)()


class MySqlConnection(object):
    def assert_fails(self, ip, user_name, password):
        try:
            with mysql.create_mysql_connection(ip, user_name, password):
                pass
            fail("Should have failed to connect: mysql --host %s -u %s -p%s"
                 % (ip, user_name, password))
        except mysql.MySqlPermissionsFailure:
            return  # Good, this is what we wanted.
        except mysql.MySqlConnectionFailure as mcf:
            fail("Expected to see permissions failure. Instead got message:"
                 "%s" % mcf.message)

    @tenacity.retry(
        wait=tenacity.wait_fixed(3),
        stop=tenacity.stop_after_attempt(5),
        reraise=True
    )
    def create(self, ip, user_name, password):
        print("Connecting mysql, host: %s, user: %s, password: %s" %
              (ip, user_name, password))

        return mysql.create_mysql_connection(ip, user_name, password)


class LocalSqlClient(object):
    """A sqlalchemy wrapper to manage transactions."""

    def __init__(self, engine, use_flush=True):
        self.engine = engine
        self.use_flush = use_flush

    def __enter__(self):
        self.conn = self.engine.connect()
        self.trans = self.conn.begin()
        return self.conn

    def __exit__(self, type, value, traceback):
        if self.trans:
            if type is not None:  # An error occurred
                self.trans.rollback()
            else:
                if self.use_flush:
                    self.conn.execute(FLUSH)
                self.trans.commit()
        self.conn.close()

    def execute(self, t, **kwargs):
        try:
            return self.conn.execute(t, kwargs)
        except Exception:
            self.trans.rollback()
            self.trans = None
            raise

    @staticmethod
    def init_engine(user, password, host):
        return create_engine("mysql+pymysql://%s:%s@%s:3306" %
                             (user, password, host),
                             pool_recycle=1800, echo=True)



#    Copyright 2013 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
from testtools.matchers import Equals
from testtools.matchers import Is
from trove.common.exception import DatabaseForUserNotInDatabaseListError
from trove.common.exception import DatabaseInitialDatabaseDuplicateError
from trove.common.exception import DatabaseInitialUserDuplicateError
from trove.extensions.mysql.common import populate_users
from trove.extensions.mysql.common import populate_validated_databases
from trove.tests.unittests import trove_testtools


class MySqlCommonTest(trove_testtools.TestCase):

    def setUp(self):
        super(MySqlCommonTest, self).setUp()

    def tearDown(self):
        super(MySqlCommonTest, self).tearDown()

    def test_initial_databases_none(self):
        databases = []
        result = populate_validated_databases(databases)
        self.assertThat(len(result), Is(0))

    def test_initial_databases_single(self):
        databases = [{'name': 'one_db'}]
        result = populate_validated_databases(databases)
        self.assertThat(len(result), Is(1))
        self.assertThat(result[0]['_name'], Equals('one_db'))

    def test_initial_databases_unique(self):
        databases = [{'name': 'one_db'}, {'name': 'diff_db'}]
        result = populate_validated_databases(databases)
        self.assertThat(len(result), Is(2))

    def test_initial_databases_duplicate(self):
        databases = [{'name': 'same_db'}, {'name': 'same_db'}]
        self.assertRaises(DatabaseInitialDatabaseDuplicateError,
                          populate_validated_databases, databases)

    def test_initial_databases_intermingled(self):
        databases = [{'name': 'a_db'}, {'name': 'b_db'}, {'name': 'a_db'}]
        self.assertRaises(DatabaseInitialDatabaseDuplicateError,
                          populate_validated_databases, databases)

    def test_populate_users_single(self):
        users = [{'name': 'bob', 'password': 'x'}]
        result = populate_users(users)
        self.assertThat(len(result), Is(1))
        self.assertThat(result[0]['_name'], Equals('bob'))
        self.assertThat(result[0]['_password'], Equals('x'))

    def test_populate_users_unique_host(self):
        users = [{'name': 'bob', 'password': 'x', 'host': '127.0.0.1'},
                 {'name': 'bob', 'password': 'x', 'host': '128.0.0.1'}]
        result = populate_users(users)
        self.assertThat(len(result), Is(2))

    def test_populate_users_unique_name(self):
        users = [{'name': 'bob', 'password': 'x', 'host': '127.0.0.1'},
                 {'name': 'tom', 'password': 'x', 'host': '127.0.0.1'}]
        result = populate_users(users)
        self.assertThat(len(result), Is(2))

    def test_populate_users_duplicate(self):
        users = [{'name': 'bob', 'password': 'x', 'host': '127.0.0.1'},
                 {'name': 'bob', 'password': 'y', 'host': '127.0.0.1'}]
        self.assertRaises(DatabaseInitialUserDuplicateError,
                          populate_users, users)

    def test_populate_unique_users_unique_host(self):
        users = [{'name': 'bob', 'password': 'x', 'host': '127.0.0.1'},
                 {'name': 'tom', 'password': 'x', 'host': '128.0.0.1'}]
        result = populate_users(users)
        self.assertThat(len(result), Is(2))

    def test_populate_users_intermingled(self):
        users = [{'name': 'bob', 'password': 'x', 'host': '127.0.0.1'},
                 {'name': 'tom', 'password': 'y', 'host': '127.0.0.1'},
                 {'name': 'bob', 'password': 'z', 'host': '127.0.0.1'},
                 {'name': 'bob', 'password': 'x', 'host': '128.0.0.1'},
                 {'name': 'tom', 'password': 'x', 'host': '128.0.0.1'}]
        self.assertRaises(DatabaseInitialUserDuplicateError,
                          populate_users, users)

    def test_populate_users_both_db_list_empty(self):
        initial_databases = []
        users = [{"name": "bob", "password": "x"}]
        result = populate_users(users, initial_databases)
        self.assertThat(len(result), Is(1))

    def test_populate_users_initial_db_list_empty(self):
        initial_databases = []
        users = [{"name": "bob", "password": "x",
                  "databases": [{"name": "my_db"}]}]
        self.assertRaises(DatabaseForUserNotInDatabaseListError,
                          populate_users, users, initial_databases)

    def test_populate_users_user_db_list_empty(self):
        initial_databases = ['my_db']
        users = [{"name": "bob", "password": "x"}]
        result = populate_users(users, initial_databases)
        self.assertThat(len(result), Is(1))

    def test_populate_users_db_in_list(self):
        initial_databases = ['my_db']
        users = [{"name": "bob", "password": "x",
                  "databases": [{"name": "my_db"}]}]
        result = populate_users(users, initial_databases)
        self.assertThat(len(result), Is(1))

    def test_populate_users_db_multi_in_list(self):
        initial_databases = ['a_db', 'b_db', 'c_db', 'd_db']
        users = [{"name": "bob", "password": "x",
                  "databases": [{"name": "a_db"}]},
                 {"name": "tom", "password": "y",
                  "databases": [{"name": "c_db"}]},
                 {"name": "sue", "password": "z",
                  "databases": [{"name": "c_db"}]}]
        result = populate_users(users, initial_databases)
        self.assertThat(len(result), Is(3))

    def test_populate_users_db_not_in_list(self):
        initial_databases = ['a_db', 'b_db', 'c_db', 'd_db']
        users = [{"name": "bob", "password": "x",
                  "databases": [{"name": "fake_db"}]}]
        self.assertRaises(DatabaseForUserNotInDatabaseListError,
                          populate_users, users, initial_databases)

    def test_populate_users_db_multi_not_in_list(self):
        initial_databases = ['a_db', 'b_db', 'c_db', 'd_db']
        users = [{"name": "bob", "password": "x",
                  "databases": [{"name": "a_db"}]},
                 {"name": "tom", "password": "y",
                  "databases": [{"name": "fake_db"}]},
                 {"name": "sue", "password": "z",
                  "databases": [{"name": "d_db"}]}]
        self.assertRaises(DatabaseForUserNotInDatabaseListError,
                          populate_users, users, initial_databases)



#    Copyright 2012 OpenStack Foundation
#    Copyright 2014 Rackspace Hosting
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg as oslo_config
from oslo_log import log as logging

from trove.common import cfg
from trove.common import configurations
from trove.common import exception
from trove.common import utils

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

ENV = utils.ENV

SERVICE_PARSERS = {
    'mongodb': configurations.MongoDBConfParser,
    'mysql': configurations.MySQLConfParser,
    'percona': configurations.MySQLConfParser,
    'mariadb': configurations.MySQLConfParser,
    'pxc': configurations.MySQLConfParser,
    'postgresql': configurations.PostgresqlConfParser,
    'cassandra': configurations.CassandraConfParser,
    'redis': configurations.RedisConfParser,
    'vertica': configurations.VerticaConfParser,
    'db2': configurations.DB2ConfParser,
}


class SingleInstanceConfigTemplate(object):
    """This class selects a single configuration file by database type for
        rendering on the guest
    """

    template_name = "config.template"

    def __init__(self, datastore_version, flavor_dict, instance_id):
        """Constructor

        :param datastore_version: The datastore version.
        :type datastore_version: DatastoreVersion
        :param flavor_dict: dict containing flavor details for use in jinja.
        :type flavor_dict: dict.
        :param instance_id: trove instance id
        :type instance_id: str

        """
        self.flavor_dict = flavor_dict
        self.datastore_version = datastore_version
        # TODO(tim.simpson): The current definition of datastore_version is a
        #                    bit iffy and I believe will change soon, so I'm
        #                    creating a dictionary here for jinja to consume
        #                    rather than pass in the datastore version object.
        self.datastore_dict = {
            'name': self.datastore_version.datastore_name,
            'manager': self.datastore_version.manager,
            'version': self.datastore_version.name,
        }
        self.instance_id = instance_id

    def get_template(self):
        patterns = ['{name}/{version}/{template_name}',
                    '{name}/{template_name}',
                    '{manager}/{template_name}']
        context = self.datastore_dict.copy()
        context['template_name'] = self.template_name
        names = [name.format(**context) for name in patterns]
        return ENV.select_template(names)

    def render(self, **kwargs):
        """Renders the jinja template

        :returns: str -- The rendered configuration file

        """
        template = self.get_template()
        server_id = self._calculate_unique_id()
        self.config_contents = template.render(
            flavor=self.flavor_dict,
            datastore=self.datastore_dict,
            server_id=server_id, **kwargs)
        return self.config_contents

    def render_dict(self):
        """
        Renders the default configuration template file as a dictionary
        to apply the default configuration dynamically.
        """
        config = self.render()
        cfg_parser = SERVICE_PARSERS.get(self.datastore_version.manager)
        if not cfg_parser:
            raise exception.NoConfigParserFound(
                datastore_manager=self.datastore_version.manager)
        return cfg_parser(config).parse()

    def _calculate_unique_id(self):
        """
        Returns a positive unique id based off of the instance id

        :return: a positive integer
        """
        return abs(hash(self.instance_id) % (2 ** 31))


def _validate_datastore(datastore_manager):
    try:
        CONF.get(datastore_manager)
    except oslo_config.NoSuchOptError:
        raise exception.InvalidDatastoreManager(
            datastore_manager=datastore_manager)


class ReplicaSourceConfigTemplate(SingleInstanceConfigTemplate):
    template_name = "replica_source.config.template"


class ReplicaConfigTemplate(SingleInstanceConfigTemplate):
    template_name = "replica.config.template"


class ClusterConfigTemplate(SingleInstanceConfigTemplate):
    template_name = "cluster.config.template"



# Copyright (c) 2015 Rackspace Hosting
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from trove.common import clients
from trove.common import exception
from trove.datastore import models as datastore_models
from trove.tests.unittests.datastore.base import TestDatastoreBase


class TestDatastoreVersionMetadata(TestDatastoreBase):
    def setUp(self):
        super(TestDatastoreVersionMetadata, self).setUp()
        self.dsmetadata = datastore_models.DatastoreVersionMetadata
        self.volume_types = [
            {'id': 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa', 'name': 'type_1'},
            {'id': 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb', 'name': 'type_2'},
            {'id': 'cccccccc-cccc-cccc-cccc-cccccccccccc', 'name': 'type_3'},
        ]

    def tearDown(self):
        super(TestDatastoreVersionMetadata, self).tearDown()

    def test_map_flavors_to_datastore(self):
        datastore = datastore_models.Datastore.load(self.ds_name)
        ds_version = datastore_models.DatastoreVersion.load(
            datastore, self.ds_version_name)
        mapping = datastore_models.DBDatastoreVersionMetadata.find_by(
            datastore_version_id=ds_version.id,
            value=self.flavor_id, deleted=False, key='flavor')
        self.assertEqual(str(self.flavor_id), mapping.value)
        self.assertEqual(ds_version.id, mapping.datastore_version_id)
        self.assertEqual('flavor', str(mapping.key))

    def test_map_volume_types_to_datastores(self):
        datastore = datastore_models.Datastore.load(self.ds_name)
        ds_version = datastore_models.DatastoreVersion.load(
            datastore, self.ds_version_name)
        mapping = datastore_models.DBDatastoreVersionMetadata.find_by(
            datastore_version_id=ds_version.id,
            value=self.volume_type, deleted=False, key='volume_type')
        self.assertEqual(str(self.volume_type), mapping.value)
        self.assertEqual(ds_version.id, mapping.datastore_version_id)
        self.assertEqual('volume_type', str(mapping.key))

    def test_add_existing_flavor_associations(self):
        dsmetadata = datastore_models.DatastoreVersionMetadata
        self.assertRaises(
            exception.DatastoreFlavorAssociationAlreadyExists,
            dsmetadata.add_datastore_version_flavor_association,
            self.test_id, [self.flavor_id])

    def test_add_existing_volume_type_associations(self):
        dsmetadata = datastore_models.DatastoreVersionMetadata
        self.assertRaises(
            exception.DatastoreVolumeTypeAssociationAlreadyExists,
            dsmetadata.add_datastore_version_volume_type_association,
            self.test_id, [self.volume_type])

    def test_delete_nonexistent_flavor_mapping(self):
        dsmeta = datastore_models.DatastoreVersionMetadata
        self.assertRaises(
            exception.DatastoreFlavorAssociationNotFound,
            dsmeta.delete_datastore_version_flavor_association,
            self.test_id, flavor_id=2)

    def test_delete_nonexistent_volume_type_mapping(self):
        dsmeta = datastore_models.DatastoreVersionMetadata
        self.assertRaises(
            exception.DatastoreVolumeTypeAssociationNotFound,
            dsmeta.delete_datastore_version_volume_type_association,
            self.test_id,
            volume_type_name='some random thing')

    def test_delete_flavor_mapping(self):
        flavor_id = 2
        dsmetadata = datastore_models.DatastoreVersionMetadata
        dsmetadata.add_datastore_version_flavor_association(
            self.test_id, [flavor_id])
        dsmetadata.delete_datastore_version_flavor_association(
            self.test_id, flavor_id)
        datastore = datastore_models.Datastore.load(self.ds_name)
        ds_version = datastore_models.DatastoreVersion.load(
            datastore,
            self.ds_version_name)
        mapping = datastore_models.DBDatastoreVersionMetadata.find_by(
            datastore_version_id=ds_version.id, value=flavor_id, key='flavor')
        self.assertTrue(mapping.deleted)
        # check update
        dsmetadata.add_datastore_version_flavor_association(
            self.test_id, [flavor_id])
        mapping = datastore_models.DBDatastoreVersionMetadata.find_by(
            datastore_version_id=ds_version.id, value=flavor_id, key='flavor')
        self.assertFalse(mapping.deleted)
        # clear the mapping
        datastore_models.DatastoreVersionMetadata. \
            delete_datastore_version_flavor_association(self.test_id,
                                                        flavor_id)

    def test_delete_volume_type_mapping(self):
        volume_type = 'this is bogus'
        dsmetadata = datastore_models.DatastoreVersionMetadata
        dsmetadata.add_datastore_version_volume_type_association(
            self.test_id, [volume_type])
        dsmetadata.delete_datastore_version_volume_type_association(
            self.test_id, volume_type)
        datastore = datastore_models.Datastore.load(self.ds_name)
        ds_version = datastore_models.DatastoreVersion.load(
            datastore,
            self.ds_version_name)
        mapping = datastore_models.DBDatastoreVersionMetadata.find_by(
            datastore_version_id=ds_version.id, value=volume_type,
            key='volume_type')
        self.assertTrue(mapping.deleted)
        # check update
        dsmetadata.add_datastore_version_volume_type_association(
            self.test_id, [volume_type])
        mapping = datastore_models.DBDatastoreVersionMetadata.find_by(
            datastore_version_id=ds_version.id, value=volume_type,
            key='volume_type')
        self.assertFalse(mapping.deleted)
        # clear the mapping
        dsmetadata.delete_datastore_version_volume_type_association(
            self.test_id, volume_type)

    @mock.patch.object(datastore_models.DatastoreVersionMetadata,
                       'datastore_version_find')
    @mock.patch.object(datastore_models.DatastoreVersionMetadata,
                       'list_datastore_version_volume_type_associations')
    @mock.patch.object(clients, 'create_cinder_client')
    def _mocked_allowed_datastore_version_volume_types(self,
                                                       trove_volume_types,
                                                       mock_cinder_client,
                                                       mock_list, *args):
        """Call this with a list of strings specifying volume types."""
        cinder_vts = []
        for vt in self.volume_types:
            cinder_type = mock.Mock()
            cinder_type.id = vt.get('id')
            cinder_type.name = vt.get('name')
            cinder_vts.append(cinder_type)
        mock_cinder_client.return_value.volume_types.list.return_value = (
            cinder_vts)

        mock_trove_list_result = mock.MagicMock()
        mock_trove_list_result.count.return_value = len(trove_volume_types)
        mock_trove_list_result.__iter__.return_value = []
        for trove_vt in trove_volume_types:
            trove_type = mock.Mock()
            trove_type.value = trove_vt
            mock_trove_list_result.__iter__.return_value.append(trove_type)
        mock_list.return_value = mock_trove_list_result

        return self.dsmetadata.allowed_datastore_version_volume_types(
            None, self.random_uuid())

    def _assert_equal_types(self, test_dict, output_obj):
        self.assertEqual(test_dict.get('id'), output_obj.id)
        self.assertEqual(test_dict.get('name'), output_obj.name)

    def test_allowed_volume_types_from_ids(self):
        id1 = self.volume_types[0].get('id')
        id2 = self.volume_types[1].get('id')
        res = self._mocked_allowed_datastore_version_volume_types([id1, id2])
        self._assert_equal_types(self.volume_types[0], res[0])
        self._assert_equal_types(self.volume_types[1], res[1])

    def test_allowed_volume_types_from_names(self):
        name1 = self.volume_types[0].get('name')
        name2 = self.volume_types[1].get('name')
        res = self._mocked_allowed_datastore_version_volume_types([name1,
                                                                   name2])
        self._assert_equal_types(self.volume_types[0], res[0])
        self._assert_equal_types(self.volume_types[1], res[1])

    def test_allowed_volume_types_no_restrictions(self):
        res = self._mocked_allowed_datastore_version_volume_types([])
        self._assert_equal_types(self.volume_types[0], res[0])
        self._assert_equal_types(self.volume_types[1], res[1])
        self._assert_equal_types(self.volume_types[2], res[2])



#    Copyright 2014 Rackspace Hosting
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from trove.common import stream_codecs


class RedisConfParser(object):

    CODEC = stream_codecs.PropertiesCodec()

    def __init__(self, config):
        self.config = config

    def parse(self):
        return self.CODEC.deserialize(self.config).items()


class MySQLConfParser(object):

    SERVER_CONF_SECTION = 'mysqld'
    CODEC = stream_codecs.IniCodec(
        default_value='1', comment_markers=('#', ';', '!'))

    def __init__(self, config):
        self.config = config

    def parse(self):
        config_dict = self.CODEC.deserialize(self.config)
        mysqld_section_dict = config_dict[self.SERVER_CONF_SECTION]
        return mysqld_section_dict.items()


class MongoDBConfParser(object):

    CODEC = stream_codecs.SafeYamlCodec(default_flow_style=False)

    def __init__(self, config):
        self.config = config

    def parse(self):
        return self.CODEC.deserialize(self.config).items()


class PostgresqlConfParser(object):

    CODEC = stream_codecs.PropertiesCodec(delimiter='=')

    def __init__(self, config):
        self.config = config

    def parse(self):
        return self.CODEC.deserialize(self.config).items()


class CassandraConfParser(object):

    CODEC = stream_codecs.SafeYamlCodec(default_flow_style=False)

    def __init__(self, config):
        self.config = config

    def parse(self):
        return self.CODEC.deserialize(self.config).items()


class VerticaConfParser(object):

    CODEC = stream_codecs.PropertiesCodec(delimiter='=')

    def __init__(self, config):
        self.config = config

    def parse(self):
        return self.CODEC.deserialize(self.config).items()


class DB2ConfParser(object):

    CODEC = stream_codecs.PropertiesCodec(delimiter='=')

    def __init__(self, config):
        self.config = config

    def parse(self):
        return self.CODEC.deserialize(self.config).items()



# Copyright 2016 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from proboscis import test

from trove.tests.scenario import groups
from trove.tests.scenario.groups.test_group import TestGroup
from trove.tests.scenario.runners import test_runners


GROUP = "scenario.instance_error_create_group"


class InstanceErrorCreateRunnerFactory(test_runners.RunnerFactory):

    _runner_ns = 'instance_error_create_runners'
    _runner_cls = 'InstanceErrorCreateRunner'


@test(depends_on_groups=[groups.INST_CREATE],
      groups=[GROUP, groups.INST_ERROR_CREATE])
class InstanceErrorCreateGroup(TestGroup):
    """Test Instance Error Create functionality."""

    def __init__(self):
        super(InstanceErrorCreateGroup, self).__init__(
            InstanceErrorCreateRunnerFactory.instance())

    @test
    def create_error_instance(self):
        """Create an instance in error state."""
        self.test_runner.run_create_error_instance()

    @test(runs_after=[create_error_instance])
    def create_error2_instance(self):
        """Create another instance in error state."""
        self.test_runner.run_create_error2_instance()


@test(depends_on_classes=[InstanceErrorCreateGroup],
      groups=[GROUP, groups.INST_ERROR_CREATE_WAIT])
class InstanceErrorCreateWaitGroup(TestGroup):
    """Test that Instance Error Create Completes."""

    def __init__(self):
        super(InstanceErrorCreateWaitGroup, self).__init__(
            InstanceErrorCreateRunnerFactory.instance())

    @test
    def wait_for_error_instances(self):
        """Wait for the error instances to fail."""
        self.test_runner.run_wait_for_error_instances()

    @test(depends_on=[wait_for_error_instances])
    def validate_error_instance(self):
        """Validate the error instance fault message."""
        self.test_runner.run_validate_error_instance()

    @test(depends_on=[wait_for_error_instances],
          runs_after=[validate_error_instance])
    def validate_error2_instance(self):
        """Validate the error2 instance fault message as admin."""
        self.test_runner.run_validate_error2_instance()


@test(depends_on_classes=[InstanceErrorCreateWaitGroup],
      groups=[GROUP, groups.INST_ERROR_DELETE])
class InstanceErrorDeleteGroup(TestGroup):
    """Test Instance Error Delete functionality."""

    def __init__(self):
        super(InstanceErrorDeleteGroup, self).__init__(
            InstanceErrorCreateRunnerFactory.instance())

    @test
    def delete_error_instances(self):
        """Delete the error instances."""
        self.test_runner.run_delete_error_instances()


@test(depends_on_classes=[InstanceErrorDeleteGroup],
      groups=[GROUP, groups.INST_ERROR_DELETE_WAIT])
class InstanceErrorDeleteWaitGroup(TestGroup):
    """Test that Instance Error Delete Completes."""

    def __init__(self):
        super(InstanceErrorDeleteWaitGroup, self).__init__(
            InstanceErrorCreateRunnerFactory.instance())

    @test
    def wait_for_error_delete(self):
        """Wait for the error instances to be gone."""
        self.test_runner.run_wait_for_error_delete()



#    Copyright (c) 2014 Rackspace Hosting
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from trove.common.exception import CapabilityNotFound
from trove.datastore.models import Capability
from trove.datastore.models import CapabilityOverride
from trove.tests.unittests.datastore.base import TestDatastoreBase


class TestCapabilities(TestDatastoreBase):
    def test_capability(self):
        cap = Capability.load(self.capability_name)
        self.assertEqual(self.capability_name, cap.name)
        self.assertEqual(self.capability_desc, cap.description)
        self.assertEqual(self.capability_enabled, cap.enabled)

    def test_ds_capability_create_disabled(self):
        self.ds_cap = CapabilityOverride.create(
            self.cap1, self.datastore_version.id, enabled=False)
        self.assertFalse(self.ds_cap.enabled)

        self.ds_cap.delete()

    def test_capability_disabled(self):
        capability = Capability.load(self.capability_name)
        capability.disable()
        self.assertFalse(capability.enabled)

        self.assertFalse(Capability.load(self.capability_name).enabled)

    def test_load_nonexistent_capability(self):
        self.assertRaises(CapabilityNotFound, Capability.load, "non-existent")



# Copyright 2015 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_service import loopingcall

from trove.tests.scenario.helpers.test_helper import DataType
from trove.tests.scenario.runners.test_runners import TestRunner


class InstanceUpgradeRunner(TestRunner):

    def __init__(self):
        super(InstanceUpgradeRunner, self).__init__()

    def run_add_test_data(self):
        host = self.get_instance_host(self.instance_info.id)
        self.test_helper.add_data(DataType.small, host)

    def run_verify_test_data(self):
        host = self.get_instance_host(self.instance_info.id)
        self.test_helper.verify_data(DataType.small, host)

    def run_remove_test_data(self):
        host = self.get_instance_host(self.instance_info.id)
        self.test_helper.remove_data(DataType.small, host)

    def run_instance_upgrade(self, expected_states=['UPGRADE', 'HEALTHY'],
                             expected_http_code=202):
        instance_id = self.instance_info.id
        self.report.log("Testing upgrade on instance: %s" % instance_id)

        target_version = self.instance_info.dbaas_datastore_version
        client = self.auth_client
        client.instances.upgrade(instance_id, target_version)
        self.assert_client_code(client, expected_http_code)
        self.assert_instance_action(instance_id, expected_states)

        def _wait_for_user_list():
            try:
                all_users = self.get_user_names(client, instance_id)
                self.report.log("Users in the db instance %s: %s" %
                                (instance_id, all_users))
            except Exception as e:
                self.report.log(
                    "Failed to list users in db instance %s(will continue), "
                    "error: %s" % (instance_id, str(e))
                )
            else:
                raise loopingcall.LoopingCallDone()

        timer = loopingcall.FixedIntervalWithTimeoutLoopingCall(
            _wait_for_user_list)
        try:
            timer.start(interval=3, timeout=120).wait()
        except loopingcall.LoopingCallTimeOut:
            self.fail("Timed out: Cannot list users in the db instance %s"
                      % instance_id)



#
# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from sqlalchemy.schema import MetaData

from trove.db.sqlalchemy.migrate_repo.schema import String
from trove.db.sqlalchemy.migrate_repo.schema import Table


meta = MetaData()


def upgrade(migrate_engine):
    meta.bind = migrate_engine
    instances = Table('instances', meta, autoload=True)
    instances.c.flavor_id.alter(String(255))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy

from trove.common.policies.base import (
    PATH_CLUSTERS, PATH_CLUSTER,
    PATH_CLUSTER_INSTANCE)

rules = [
    policy.DocumentedRuleDefault(
        name='cluster:create',
        check_str='rule:admin_or_owner',
        description='Create a cluster.',
        operations=[
            {
                'path': PATH_CLUSTERS,
                'method': 'POST'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='cluster:delete',
        check_str='rule:admin_or_owner',
        description='Delete a cluster.',
        operations=[
            {
                'path': PATH_CLUSTER,
                'method': 'DELETE'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='cluster:force_delete',
        check_str='rule:admin_or_owner',
        description='Forcibly delete a cluster.',
        operations=[
            {
                'path': PATH_CLUSTER + ' (reset-status)',
                'method': 'POST'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='cluster:index',
        check_str='rule:admin_or_owner',
        description='List all clusters',
        operations=[
            {
                'path': PATH_CLUSTERS,
                'method': 'GET'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='cluster:show',
        check_str='rule:admin_or_owner',
        description='Get informations of a cluster.',
        operations=[
            {
                'path': PATH_CLUSTER,
                'method': 'GET'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='cluster:show_instance',
        check_str='rule:admin_or_owner',
        description='Get informations of a instance in a cluster.',
        operations=[
            {
                'path': PATH_CLUSTER_INSTANCE,
                'method': 'GET'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='cluster:action',
        check_str='rule:admin_or_owner',
        description='Commit an action against a cluster',
        operations=[
            {
                'path': PATH_CLUSTER,
                'method': 'POST'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='cluster:reset-status',
        check_str='rule:admin',
        description='Reset the status of a cluster to NONE.',
        operations=[
            {
                'path': PATH_CLUSTER + ' (reset-status)',
                'method': 'POST'
            }
        ])
]


def list_rules():
    return rules



# Copyright 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from novaclient.exceptions import BadRequest
from novaclient.v2.servers import Server
from oslo_messaging._drivers.common import RPCException
from proboscis import test
from testtools import TestCase

from trove.common import template
from trove.common import utils
from trove.common.exception import PollTimeOut
from trove.common.exception import TroveError
from trove.datastore.models import DatastoreVersion
from trove.guestagent import api as guest
from trove.instance import service_status as srvstatus
from trove.instance.models import DBInstance
from trove.instance.models import InstanceServiceStatus
from trove.instance.tasks import InstanceTasks
from trove.taskmanager import models
from trove.tests.fakes import nova
from trove.tests.unittests import trove_testtools
from trove.tests.util import test_config

GROUP = 'dbaas.api.instances.resize'

OLD_FLAVOR_ID = 1
NEW_FLAVOR_ID = 2
OLD_FLAVOR = nova.FLAVORS.get(OLD_FLAVOR_ID)
NEW_FLAVOR = nova.FLAVORS.get(NEW_FLAVOR_ID)


class ResizeTestBase(TestCase):

    def _init(self):
        self.instance_id = 500
        context = trove_testtools.TroveTestContext(self)
        self.db_info = DBInstance.create(
            name="instance",
            flavor_id=OLD_FLAVOR_ID,
            tenant_id=999,
            volume_size=None,
            datastore_version_id=test_config.dbaas_datastore_version_id,
            task_status=InstanceTasks.RESIZING)
        self.server = mock.MagicMock(spec=Server)
        self.instance = models.BuiltInstanceTasks(
            context,
            self.db_info,
            self.server,
            datastore_status=InstanceServiceStatus.create(
                instance_id=self.db_info.id,
                status=srvstatus.ServiceStatuses.RUNNING))
        self.instance.server.flavor = {'id': OLD_FLAVOR_ID}
        self.guest = mock.MagicMock(spec=guest.API)
        self.instance._guest = self.guest
        self.instance.refresh_compute_server_info = lambda: None
        self.instance._refresh_datastore_status = lambda: None
        self.instance.update_db = mock.Mock()
        self.instance.set_datastore_status_to_paused = mock.Mock()
        self.poll_until_side_effects = []
        self.action = None

    def tearDown(self):
        super(ResizeTestBase, self).tearDown()
        self.db_info.delete()

    def _poll_until(self, *args, **kwargs):
        try:
            effect = self.poll_until_side_effects.pop(0)
        except IndexError:
            effect = None

        if isinstance(effect, Exception):
            raise effect
        elif effect is not None:
            new_status, new_flavor_id = effect
            self.server.status = new_status
            self.instance.server.flavor['id'] = new_flavor_id

    def _datastore_changes_to(self, new_status):
        self.instance.datastore_status.status = new_status


@test(groups=[GROUP, GROUP + '.resize'])
class ResizeTests(ResizeTestBase):

    def setUp(self):
        super(ResizeTests, self).setUp()
        self._init()
        # By the time flavor objects pass over amqp to the
        # resize action they have been turned into dicts
        self.action = models.ResizeAction(self.instance,
                                          OLD_FLAVOR.__dict__,
                                          NEW_FLAVOR.__dict__)

    def _start_mysql(self):
        datastore = mock.Mock(spec=DatastoreVersion)
        datastore.datastore_name = 'mysql'
        datastore.name = 'mysql-5.7'
        datastore.manager = 'mysql'
        config = template.SingleInstanceConfigTemplate(
            datastore, NEW_FLAVOR.__dict__, self.instance.id)
        self.instance.guest.start_db_with_conf_changes(config.render(),
                                                       datastore.name)

    def test_guest_wont_stop_mysql(self):
        self.guest.stop_db.side_effect = RPCException("Could not stop MySQL!")
        self.assertRaises(RPCException, self.action.execute)
        self.assertEqual(1, self.guest.stop_db.call_count)
        self.instance.update_db.assert_called_once_with(
            task_status=InstanceTasks.NONE)

    def test_nova_wont_resize(self):
        self._datastore_changes_to(srvstatus.ServiceStatuses.SHUTDOWN)
        self.server.resize.side_effect = BadRequest(400)
        self.server.status = "ACTIVE"
        self.assertRaises(BadRequest, self.action.execute)
        self.assertEqual(1, self.guest.stop_db.call_count)
        self.server.resize.assert_called_once_with(NEW_FLAVOR_ID)
        self.guest.restart.assert_called_once()
        self.instance.update_db.assert_called_once_with(
            task_status=InstanceTasks.NONE)

    def test_nova_resize_timeout(self):
        self._datastore_changes_to(srvstatus.ServiceStatuses.SHUTDOWN)
        self.server.status = "ACTIVE"

        with mock.patch.object(utils, 'poll_until') as mock_poll_until:
            mock_poll_until.side_effect = [None, PollTimeOut()]
            self.assertRaises(PollTimeOut, self.action.execute)
            expected_calls = [
                mock.call(mock.ANY, sleep_time=2, time_out=120)] * 2
            self.assertEqual(expected_calls, mock_poll_until.call_args_list)
            self.assertEqual(1, self.guest.stop_db.call_count)
            self.server.resize.assert_called_once_with(NEW_FLAVOR_ID)
            self.instance.update_db.assert_called_once_with(
                task_status=InstanceTasks.NONE)

    def test_nova_doesnt_change_flavor(self):
        self._datastore_changes_to(srvstatus.ServiceStatuses.SHUTDOWN)

        with mock.patch.object(utils, 'poll_until') as mock_poll_until:
            self.poll_until_side_effects.extend([
                ("VERIFY_RESIZE", OLD_FLAVOR_ID),
                None,
                ("ACTIVE", OLD_FLAVOR_ID)])
            mock_poll_until.side_effect = self._poll_until

            self.assertRaisesRegex(TroveError,
                                   "flavor_id=.* and not .*",
                                   self.action.execute)
            expected_calls = [
                mock.call(mock.ANY, sleep_time=2, time_out=120)] * 3
            self.assertEqual(expected_calls, mock_poll_until.call_args_list)
            # Make sure self.poll_until_side_effects is empty
            self.assertFalse(self.poll_until_side_effects)
            self.assertEqual(1, self.guest.stop_db.call_count)
            self.server.resize.assert_called_once_with(NEW_FLAVOR_ID)
            self.instance.guest.reset_configuration.assert_called_once_with(
                mock.ANY)
            self.instance.server.revert_resize.assert_called_once()
            self.guest.restart.assert_called_once()
            self.instance.update_db.assert_called_once_with(
                task_status=InstanceTasks.NONE)

    def test_nova_resize_fails(self):
        self._datastore_changes_to(srvstatus.ServiceStatuses.SHUTDOWN)

        with mock.patch.object(utils, 'poll_until') as mock_poll_until:
            self.poll_until_side_effects.extend([
                None,
                ("ERROR", OLD_FLAVOR_ID)])
            mock_poll_until.side_effect = self._poll_until

            self.assertRaisesRegex(TroveError,
                                   "status=ERROR and not VERIFY_RESIZE",
                                   self.action.execute)

            expected_calls = [
                mock.call(mock.ANY, sleep_time=2, time_out=120)] * 2
            self.assertEqual(expected_calls, mock_poll_until.call_args_list)
            # Make sure self.poll_until_side_effects is empty
            self.assertFalse(self.poll_until_side_effects)
            self.assertEqual(1, self.guest.stop_db.call_count)
            self.server.resize.assert_called_once_with(NEW_FLAVOR_ID)
            self.instance.update_db.assert_called_once_with(
                task_status=InstanceTasks.NONE)

    def test_nova_resizes_in_weird_state(self):
        self._datastore_changes_to(srvstatus.ServiceStatuses.SHUTDOWN)

        with mock.patch.object(utils, 'poll_until') as mock_poll_until:
            self.poll_until_side_effects.extend([
                None,
                ("ACTIVE", NEW_FLAVOR_ID)])
            mock_poll_until.side_effect = self._poll_until

            self.assertRaisesRegex(TroveError,
                                   "status=ACTIVE and not VERIFY_RESIZE",
                                   self.action.execute)

            expected_calls = [
                mock.call(mock.ANY, sleep_time=2, time_out=120)] * 2
            self.assertEqual(expected_calls, mock_poll_until.call_args_list)
            # Make sure self.poll_until_side_effects is empty
            self.assertFalse(self.poll_until_side_effects)
            self.assertEqual(1, self.guest.stop_db.call_count)
            self.server.resize.assert_called_once_with(NEW_FLAVOR_ID)
            self.guest.restart.assert_called_once()
            self.instance.update_db.assert_called_once_with(
                task_status=InstanceTasks.NONE)

    def test_guest_is_not_okay(self):
        self._datastore_changes_to(srvstatus.ServiceStatuses.SHUTDOWN)

        with mock.patch.object(utils, 'poll_until') as mock_poll_until:
            self.poll_until_side_effects.extend([
                None,
                ("VERIFY_RESIZE", NEW_FLAVOR_ID),
                None,
                PollTimeOut(),
                ("ACTIVE", OLD_FLAVOR_ID)])
            mock_poll_until.side_effect = self._poll_until

            self.instance.set_datastore_status_to_paused.side_effect = (
                lambda: self._datastore_changes_to(
                    srvstatus.ServiceStatuses.PAUSED))

            self.assertRaises(PollTimeOut, self.action.execute)

            expected_calls = [
                mock.call(mock.ANY, sleep_time=2, time_out=120)] * 5
            self.assertEqual(expected_calls, mock_poll_until.call_args_list)
            # Make sure self.poll_until_side_effects is empty
            self.assertFalse(self.poll_until_side_effects)
            self.assertEqual(1, self.guest.stop_db.call_count)
            self.server.resize.assert_called_once_with(NEW_FLAVOR_ID)
            self.instance.set_datastore_status_to_paused.assert_called_once()
            self.instance.guest.reset_configuration.assert_called_once_with(
                mock.ANY)
            self.instance.server.revert_resize.assert_called_once()
            self.guest.restart.assert_called_once()
            self.instance.update_db.assert_called_once_with(
                task_status=InstanceTasks.NONE)

    def test_mysql_is_not_okay(self):
        self._datastore_changes_to(srvstatus.ServiceStatuses.SHUTDOWN)

        with mock.patch.object(utils, 'poll_until') as mock_poll_until:
            self.poll_until_side_effects.extend([
                None,
                ("VERIFY_RESIZE", NEW_FLAVOR_ID),
                PollTimeOut(),
                ("ACTIVE", OLD_FLAVOR_ID)])
            mock_poll_until.side_effect = self._poll_until

            self.instance.set_datastore_status_to_paused.side_effect = (
                lambda: self._datastore_changes_to(
                    srvstatus.ServiceStatuses.SHUTDOWN))

            self._start_mysql()
            self.assertRaises(PollTimeOut, self.action.execute)

            expected_calls = [
                mock.call(mock.ANY, sleep_time=2, time_out=120)] * 4
            self.assertEqual(expected_calls, mock_poll_until.call_args_list)
            # Make sure self.poll_until_side_effects is empty
            self.assertFalse(self.poll_until_side_effects)
            self.assertEqual(1, self.guest.stop_db.call_count)
            self.server.resize.assert_called_once_with(NEW_FLAVOR_ID)
            self.instance.set_datastore_status_to_paused.assert_called_once()
            self.instance.guest.reset_configuration.assert_called_once_with(
                mock.ANY)
            self.instance.server.revert_resize.assert_called_once()
            self.guest.restart.assert_called_once()
            self.instance.update_db.assert_called_once_with(
                task_status=InstanceTasks.NONE)

    def test_confirm_resize_fails(self):
        self._datastore_changes_to(srvstatus.ServiceStatuses.SHUTDOWN)

        with mock.patch.object(utils, 'poll_until') as mock_poll_until:
            self.poll_until_side_effects.extend([
                None,
                ("VERIFY_RESIZE", NEW_FLAVOR_ID),
                None,
                None,
                ("SHUTDOWN", NEW_FLAVOR_ID)])
            mock_poll_until.side_effect = self._poll_until

            self.instance.set_datastore_status_to_paused.side_effect = (
                lambda: self._datastore_changes_to(
                    srvstatus.ServiceStatuses.RUNNING))
            self.server.confirm_resize.side_effect = BadRequest(400)

            self._start_mysql()
            self.assertRaises(BadRequest, self.action.execute)

            expected_calls = [
                mock.call(mock.ANY, sleep_time=2, time_out=120)] * 5
            self.assertEqual(expected_calls, mock_poll_until.call_args_list)
            # Make sure self.poll_until_side_effects is empty
            self.assertFalse(self.poll_until_side_effects)
            self.assertEqual(1, self.guest.stop_db.call_count)
            self.server.resize.assert_called_once_with(NEW_FLAVOR_ID)
            self.instance.set_datastore_status_to_paused.assert_called_once()
            self.instance.server.confirm_resize.assert_called_once()
            self.instance.update_db.assert_called_once_with(
                task_status=InstanceTasks.NONE)

    def test_revert_nova_fails(self):
        self._datastore_changes_to(srvstatus.ServiceStatuses.SHUTDOWN)

        with mock.patch.object(utils, 'poll_until') as mock_poll_until:
            self.poll_until_side_effects.extend([
                None,
                ("VERIFY_RESIZE", NEW_FLAVOR_ID),
                None,
                PollTimeOut(),
                ("ERROR", OLD_FLAVOR_ID)])
            mock_poll_until.side_effect = self._poll_until

            self.instance.set_datastore_status_to_paused.side_effect = (
                lambda: self._datastore_changes_to(
                    srvstatus.ServiceStatuses.PAUSED))

            self.assertRaises(PollTimeOut, self.action.execute)

            expected_calls = [
                mock.call(mock.ANY, sleep_time=2, time_out=120)] * 5
            self.assertEqual(expected_calls, mock_poll_until.call_args_list)
            # Make sure self.poll_until_side_effects is empty
            self.assertFalse(self.poll_until_side_effects)
            self.assertEqual(1, self.guest.stop_db.call_count)
            self.server.resize.assert_called_once_with(NEW_FLAVOR_ID)
            self.instance.set_datastore_status_to_paused.assert_called_once()
            self.instance.guest.reset_configuration.assert_called_once_with(
                mock.ANY)
            self.instance.server.revert_resize.assert_called_once()
            self.instance.update_db.assert_called_once_with(
                task_status=InstanceTasks.NONE)


@test(groups=[GROUP, GROUP + '.migrate'])
class MigrateTests(ResizeTestBase):

    def setUp(self):
        super(MigrateTests, self).setUp()
        self._init()
        self.action = models.MigrateAction(self.instance)

    def test_successful_migrate(self):
        self._datastore_changes_to(srvstatus.ServiceStatuses.SHUTDOWN)

        with mock.patch.object(utils, 'poll_until') as mock_poll_until:
            self.poll_until_side_effects.extend([
                None,
                ("VERIFY_RESIZE", NEW_FLAVOR_ID),
                None,
                None])
            mock_poll_until.side_effect = self._poll_until

            self.instance.set_datastore_status_to_paused.side_effect = (
                lambda: self._datastore_changes_to(
                    srvstatus.ServiceStatuses.RUNNING))

            self.action.execute()

            expected_calls = [
                mock.call(mock.ANY, sleep_time=2, time_out=120)] * 4
            self.assertEqual(expected_calls, mock_poll_until.call_args_list)
            # Make sure self.poll_until_side_effects is empty
            self.assertFalse(self.poll_until_side_effects)
            self.assertEqual(1, self.guest.stop_db.call_count)
            self.server.migrate.assert_called_once_with(force_host=None)
            self.instance.set_datastore_status_to_paused.assert_called_once()
            self.instance.server.confirm_resize.assert_called_once()
            self.instance.update_db.assert_called_once_with(
                task_status=InstanceTasks.NONE)



# Copyright 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from trove.common import timeutils


class LimitView(object):

    def __init__(self, rate_limit):
        self.rate_limit = rate_limit

    def data(self):
        get_utc = datetime.datetime.utcfromtimestamp
        next_avail = get_utc(self.rate_limit.get("resetTime", 0))

        return {"limit": {
            "nextAvailable": timeutils.isotime(next_avail),
            "remaining": self.rate_limit.get("remaining", 0),
            "unit": self.rate_limit.get("unit", ""),
            "value": self.rate_limit.get("value", ""),
            "verb": self.rate_limit.get("verb", ""),
            "uri": self.rate_limit.get("URI", ""),
            "regex": self.rate_limit.get("regex", "")
        }
        }


class LimitViews(object):

    def __init__(self, abs_limits, rate_limits):
        self.abs_limits = abs_limits
        self.rate_limits = rate_limits

    def data(self):
        data = []
        abs_view = dict()
        abs_view["verb"] = "ABSOLUTE"
        for resource_name, abs_limit in self.abs_limits.items():
            abs_view["max_" + resource_name] = abs_limit

        data.append(abs_view)
        for limit in self.rate_limits:
            data.append(LimitView(limit).data()["limit"])
        return {"limits": data}



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from lxml import etree
from oslo_log import log as logging
from oslo_utils import encodeutils
import routes
import stevedore
import webob.dec
import webob.exc

from trove.common import base_exception as exception
from trove.common import base_wsgi
from trove.common.i18n import _
from trove.common import wsgi

LOG = logging.getLogger(__name__)
DEFAULT_XMLNS = "http://docs.openstack.org/trove"
XMLNS_ATOM = "http://www.w3.org/2005/Atom"


class ExtensionDescriptor(object, metaclass=abc.ABCMeta):
    """Base class that defines the contract for extensions.

    Note that you don't have to derive from this class to have a valid
    extension; it is purely a convenience.

    """
    @abc.abstractmethod
    def get_name(self):
        """The name of the extension.

        e.g. 'Fox In Socks'

        """
        pass

    @abc.abstractmethod
    def get_alias(self):
        """The alias for the extension.

        e.g. 'FOXNSOX'

        """
        pass

    @abc.abstractmethod
    def get_description(self):
        """Friendly description for the extension.

        e.g. 'The Fox In Socks Extension'

        """
        pass

    @abc.abstractmethod
    def get_namespace(self):
        """The XML namespace for the extension.

        e.g. 'http://www.fox.in.socks/api/ext/pie/v1.0'

        """
        pass

    @abc.abstractmethod
    def get_updated(self):
        """The timestamp when the extension was last updated.

        e.g. '2011-01-22T13:25:27-06:00'

        """
        pass

    def get_resources(self):
        """List of extensions.ResourceExtension extension objects.

        Resources define new nouns, and are accessible through URLs.

        """
        resources = []
        return resources

    def get_actions(self):
        """List of extensions.ActionExtension extension objects.

        Actions are verbs callable from the API.

        """
        actions = []
        return actions

    def get_request_extensions(self):
        """List of extensions.RequestException extension objects.

        Request extensions are used to handle custom request data.

        """
        request_exts = []
        return request_exts


class ActionExtensionController(object):
    def __init__(self, application):
        self.application = application
        self.action_handlers = {}

    def add_action(self, action_name, handler):
        self.action_handlers[action_name] = handler

    def action(self, req, id, body):
        for action_name, handler in self.action_handlers.items():
            if action_name in body:
                return handler(body, req, id)
        # no action handler found (bump to downstream application)
        res = self.application
        return res


class ActionExtensionResource(wsgi.Resource):

    def __init__(self, application):
        controller = ActionExtensionController(application)
        wsgi.Resource.__init__(self, controller)

    def add_action(self, action_name, handler):
        self.controller.add_action(action_name, handler)


class RequestExtensionController(object):

    def __init__(self, application):
        self.application = application
        self.handlers = []

    def add_handler(self, handler):
        self.handlers.append(handler)

    def process(self, req, *args, **kwargs):
        res = req.get_response(self.application)
        # currently request handlers are un-ordered
        for handler in self.handlers:
            res = handler(req, res)
        return res


class RequestExtensionResource(wsgi.Resource):

    def __init__(self, application):
        controller = RequestExtensionController(application)
        wsgi.Resource.__init__(self, controller)

    def add_handler(self, handler):
        self.controller.add_handler(handler)


class ExtensionsResource(wsgi.Resource):

    def __init__(self, extension_manager):
        self.extension_manager = extension_manager
        body_serializers = {'application/xml': ExtensionsXMLSerializer()}
        serializer = base_wsgi.ResponseSerializer(
            body_serializers=body_serializers)
        super(ExtensionsResource, self).__init__(self, None, serializer)

    def _translate(self, ext):
        ext_data = {}
        ext_data['name'] = ext.get_name()
        ext_data['alias'] = ext.get_alias()
        ext_data['description'] = ext.get_description()
        ext_data['namespace'] = ext.get_namespace()
        ext_data['updated'] = ext.get_updated()
        ext_data['links'] = []
        return ext_data

    def index(self, req):
        extensions = []
        for _alias, ext in self.extension_manager.extensions.items():
            extensions.append(self._translate(ext))
        return dict(extensions=extensions)

    def show(self, req, id):
        # NOTE(dprince): the extensions alias is used as the 'id' for show
        ext = self.extension_manager.extensions.get(id, None)
        if not ext:
            raise webob.exc.HTTPNotFound(
                _("Extension with alias %s does not exist") % id)

        return dict(extension=self._translate(ext))

    def delete(self, req, id):
        raise webob.exc.HTTPNotFound()

    def create(self, req):
        raise webob.exc.HTTPNotFound()


class ExtensionMiddleware(wsgi.Middleware):
    """Extensions middleware for WSGI."""

    @classmethod
    def factory(cls, global_config, **local_config):
        """Paste factory."""
        def _factory(app):
            return cls(app, global_config, **local_config)
        return _factory

    def _action_ext_resources(self, application, ext_mgr, mapper):
        """Return a dict of ActionExtensionResource-s by collection."""
        action_resources = {}
        for action in ext_mgr.get_actions():
            if action.collection not in action_resources.keys():
                resource = ActionExtensionResource(application)
                mapper.connect("/%s/:(id)/action.:(format)" %
                               action.collection,
                               action='action',
                               controller=resource,
                               conditions=dict(method=['POST']))
                mapper.connect("/%s/:(id)/action" %
                               action.collection,
                               action='action',
                               controller=resource,
                               conditions=dict(method=['POST']))
                action_resources[action.collection] = resource

        return action_resources

    def _request_ext_resources(self, application, ext_mgr, mapper):
        """Returns a dict of RequestExtensionResource-s by collection."""
        request_ext_resources = {}
        for req_ext in ext_mgr.get_request_extensions():
            if req_ext.key not in request_ext_resources.keys():
                resource = RequestExtensionResource(application)
                mapper.connect(req_ext.url_route + '.:(format)',
                               action='process',
                               controller=resource,
                               conditions=req_ext.conditions)

                mapper.connect(req_ext.url_route,
                               action='process',
                               controller=resource,
                               conditions=req_ext.conditions)
                request_ext_resources[req_ext.key] = resource

        return request_ext_resources

    def __init__(self, application, config, ext_mgr=None):
        ext_mgr = (ext_mgr or
                   ExtensionManager())
        mapper = routes.Mapper()

        # extended resources
        for resource_ext in ext_mgr.get_resources():
            LOG.debug('Extended resource: %s', resource_ext.collection)
            controller_resource = wsgi.Resource(resource_ext.controller,
                                                resource_ext.deserializer,
                                                resource_ext.serializer)
            self._map_custom_collection_actions(resource_ext, mapper,
                                                controller_resource)
            kargs = dict(controller=controller_resource,
                         collection=resource_ext.collection_actions,
                         member=resource_ext.member_actions)
            if resource_ext.parent:
                kargs['parent_resource'] = resource_ext.parent
            mapper.resource(resource_ext.collection,
                            resource_ext.collection, **kargs)

        # extended actions
        action_resources = self._action_ext_resources(application, ext_mgr,
                                                      mapper)
        for action in ext_mgr.get_actions():
            LOG.debug('Extended action: %s', action.action_name)
            resource = action_resources[action.collection]
            resource.add_action(action.action_name, action.handler)

        # extended requests
        req_controllers = self._request_ext_resources(application, ext_mgr,
                                                      mapper)
        for request_ext in ext_mgr.get_request_extensions():
            LOG.debug('Extended request: %s', request_ext.key)
            controller = req_controllers[request_ext.key]
            controller.add_handler(request_ext.handler)

        self._router = routes.middleware.RoutesMiddleware(self._dispatch,
                                                          mapper)

        super(ExtensionMiddleware, self).__init__(application)

    def _map_custom_collection_actions(self, resource_ext, mapper,
                                       controller_resource):
        for action, method in resource_ext.collection_actions.items():
            parent = resource_ext.parent
            conditions = dict(method=[method])
            path = "/%s/%s" % (resource_ext.collection, action)

            path_prefix = ""
            if parent:
                path_prefix = "/%s/{%s_id}" % (parent["collection_name"],
                                               parent["member_name"])

            with mapper.submapper(controller=controller_resource,
                                  action=action,
                                  path_prefix=path_prefix,
                                  conditions=conditions) as submap:
                submap.connect(path_prefix + path, path)
                submap.connect(path_prefix + path + "_format",
                               "%s.:(format)" % path)

    @webob.dec.wsgify(RequestClass=wsgi.Request)
    def __call__(self, req):
        """Route the incoming request with router."""
        req.environ['extended.app'] = self.application
        return self._router

    @staticmethod
    @webob.dec.wsgify(RequestClass=wsgi.Request)
    def _dispatch(req):
        """Dispatch the request.

        Returns the routed WSGI app's response or defers to the extended
        application.

        """
        match = req.environ['wsgiorg.routing_args'][1]
        if not match:
            return req.environ['extended.app']
        app = match['controller']
        return app


class ExtensionManager(object):

    EXT_NAMESPACE = 'trove.api.extensions'

    def __init__(self):
        LOG.debug('Initializing extension manager.')

        self.extensions = {}
        self._load_all_extensions()

    def get_resources(self):
        """Returns a list of ResourceExtension objects."""
        resources = []
        extension_resource = ExtensionsResource(self)
        res_ext = ResourceExtension('extensions',
                                    extension_resource,
                                    serializer=extension_resource.serializer)
        resources.append(res_ext)
        for alias, ext in self.extensions.items():
            try:
                resources.extend(ext.get_resources())
            except AttributeError:
                pass
        return resources

    def get_actions(self):
        """Returns a list of ActionExtension objects."""
        actions = []
        for alias, ext in self.extensions.items():
            try:
                actions.extend(ext.get_actions())
            except AttributeError:
                pass
        return actions

    def get_request_extensions(self):
        """Returns a list of RequestExtension objects."""
        request_exts = []
        for alias, ext in self.extensions.items():
            try:
                request_exts.extend(ext.get_request_extensions())
            except AttributeError:
                pass
        return request_exts

    def _check_extension(self, extension):
        """Checks for required methods in extension objects."""
        try:
            LOG.debug('Ext name: %s', extension.get_name())
            LOG.debug('Ext alias: %s', extension.get_alias())
            LOG.debug('Ext description: %s', extension.get_description())
            LOG.debug('Ext namespace: %s', extension.get_namespace())
            LOG.debug('Ext updated: %s', extension.get_updated())
        except AttributeError as ex:
            LOG.exception("Exception loading extension: %s",
                          encodeutils.exception_to_unicode(ex))
            return False
        return True

    def _check_load_extension(self, ext):
        LOG.debug('Ext: %s', ext.obj)
        return isinstance(ext.obj, ExtensionDescriptor)

    def _load_all_extensions(self):
        self.api_extension_manager = stevedore.enabled.EnabledExtensionManager(
            namespace=self.EXT_NAMESPACE,
            check_func=self._check_load_extension,
            invoke_on_load=True,
            invoke_kwds={})
        self.api_extension_manager.map(self.add_extension)

    def add_extension(self, ext):
        ext = ext.obj
        # Do nothing if the extension doesn't check out
        if not self._check_extension(ext):
            return

        alias = ext.get_alias()
        LOG.debug('Loaded extension: %s', alias)

        if alias in self.extensions:
            raise exception.Error("Found duplicate extension: %s" % alias)
        self.extensions[alias] = ext


class RequestExtension(object):

    def __init__(self, method, url_route, handler):
        self.url_route = url_route
        self.handler = handler
        self.conditions = dict(method=[method])
        self.key = "%s-%s" % (method, url_route)


class ActionExtension(object):

    def __init__(self, collection, action_name, handler):
        self.collection = collection
        self.action_name = action_name
        self.handler = handler


class BaseResourceExtension(object):

    def __init__(self, collection, controller, parent=None,
                 collection_actions=None, member_actions=None,
                 deserializer=None, serializer=None):
        if not collection_actions:
            collection_actions = {}
        if not member_actions:
            member_actions = {}
        self.collection = collection
        self.controller = controller
        self.parent = parent
        self.collection_actions = collection_actions
        self.member_actions = member_actions
        self.deserializer = deserializer
        self.serializer = serializer


class ExtensionsXMLSerializer(base_wsgi.XMLDictSerializer):

    def __init__(self):
        self.nsmap = {None: DEFAULT_XMLNS, 'atom': XMLNS_ATOM}

    def show(self, ext_dict):
        ext = etree.Element('extension', nsmap=self.nsmap)
        self._populate_ext(ext, ext_dict['extension'])
        return self._to_xml(ext)

    def index(self, exts_dict):
        exts = etree.Element('extensions', nsmap=self.nsmap)
        for ext_dict in exts_dict['extensions']:
            ext = etree.SubElement(exts, 'extension')
            self._populate_ext(ext, ext_dict)
        return self._to_xml(exts)

    def _populate_ext(self, ext_elem, ext_dict):
        """Populate an extension xml element from a dict."""

        ext_elem.set('name', ext_dict['name'])
        ext_elem.set('namespace', ext_dict['namespace'])
        ext_elem.set('alias', ext_dict['alias'])
        ext_elem.set('updated', ext_dict['updated'])
        desc = etree.Element('description')
        desc.text = ext_dict['description']
        ext_elem.append(desc)
        for link in ext_dict.get('links', []):
            elem = etree.SubElement(ext_elem, '{%s}link' % XMLNS_ATOM)
            elem.set('rel', link['rel'])
            elem.set('href', link['href'])
            elem.set('type', link['type'])
        return ext_elem

    def _to_xml(self, root):
        """Convert the xml object to an xml string."""

        return etree.tostring(root, encoding='UTF-8')


class ResourceExtension(BaseResourceExtension):
    def __init__(self, collection, controller, parent=None,
                 collection_actions=None, member_actions=None,
                 deserializer=None, serializer=None):
        super(ResourceExtension, self).__init__(
            collection, controller,
            parent=parent,
            collection_actions=collection_actions,
            member_actions=member_actions,
            deserializer=wsgi.RequestDeserializer(),
            serializer=wsgi.TroveResponseSerializer())


class TroveExtensionMiddleware(ExtensionMiddleware):

    def __init__(self, application, ext_mgr=None):
        ext_mgr = (ext_mgr or
                   ExtensionManager())
        mapper = routes.Mapper()

        # extended resources
        for resource_ext in ext_mgr.get_resources():
            LOG.debug('Extended resource: %s', resource_ext.collection)
            # The only difference here is that we are using our common
            # wsgi.Resource instead of the openstack common wsgi.Resource
            exception_map = None
            if hasattr(resource_ext.controller, 'exception_map'):
                exception_map = resource_ext.controller.exception_map
            controller_resource = wsgi.Resource(resource_ext.controller,
                                                resource_ext.deserializer,
                                                resource_ext.serializer,
                                                exception_map)

            self._map_custom_collection_actions(resource_ext, mapper,
                                                controller_resource)
            kargs = dict(controller=controller_resource,
                         collection=resource_ext.collection_actions,
                         member=resource_ext.member_actions)
            if resource_ext.parent:
                kargs['parent_resource'] = resource_ext.parent
            mapper.resource(resource_ext.collection,
                            resource_ext.collection, **kargs)

            mapper.connect(("/%s/{id}" % resource_ext.collection),
                           controller=controller_resource,
                           action='edit',
                           conditions={'method': ['PATCH']})

        # extended actions
        action_resources = self._action_ext_resources(application, ext_mgr,
                                                      mapper)
        for action in ext_mgr.get_actions():
            LOG.debug('Extended action: %s', action.action_name)
            resource = action_resources[action.collection]
            resource.add_action(action.action_name, action.handler)

        # extended requests
        req_controllers = self._request_ext_resources(application, ext_mgr,
                                                      mapper)
        for request_ext in ext_mgr.get_request_extensions():
            LOG.debug('Extended request: %s', request_ext.key)
            controller = req_controllers[request_ext.key]
            controller.add_handler(request_ext.handler)

        self._router = routes.middleware.RoutesMiddleware(self._dispatch,
                                                          mapper)

        super(ExtensionMiddleware, self).__init__(application)


def factory(global_config, **local_config):
    """Paste factory."""
    def _factory(app):
        ext_mgr = ExtensionManager()
        return TroveExtensionMiddleware(app, ext_mgr)
    return _factory



# Copyright 2015 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from proboscis import test

from trove.tests.scenario import groups
from trove.tests.scenario.groups.test_group import TestGroup
from trove.tests.scenario.runners import test_runners


GROUP = "scenario.instance_delete_group"


class InstanceDeleteRunnerFactory(test_runners.RunnerFactory):

    _runner_ns = 'instance_delete_runners'
    _runner_cls = 'InstanceDeleteRunner'


@test(depends_on_groups=[groups.INST_CREATE],
      groups=[GROUP, groups.INST_DELETE],
      runs_after_groups=[groups.USER_ACTION_INST_DELETE_WAIT,
                         groups.REPL_INST_DELETE_WAIT])
class InstanceDeleteGroup(TestGroup):
    """Test Instance Delete functionality."""

    def __init__(self):
        super(InstanceDeleteGroup, self).__init__(
            InstanceDeleteRunnerFactory.instance())

    @test
    def instance_delete(self):
        """Delete an existing instance."""
        self.test_runner.run_instance_delete()


@test(depends_on_classes=[InstanceDeleteGroup],
      groups=[GROUP, groups.INST_DELETE_WAIT])
class InstanceDeleteWaitGroup(TestGroup):
    """Test that Instance Delete Completes."""

    def __init__(self):
        super(InstanceDeleteWaitGroup, self).__init__(
            InstanceDeleteRunnerFactory.instance())

    @test
    def instance_delete_wait(self):
        """Wait for existing instance to be gone."""
        self.test_runner.run_instance_delete_wait()



# Copyright 2020 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import psycopg2


class PostgresConnection(object):
    def __init__(self, user, password='', host='/var/run/postgresql',
                 port=5432):
        """Utility class to communicate with PostgreSQL.

        Connect with socket rather than IP or localhost address to avoid
        manipulation of pg_hba.conf when the database is running inside
        container with bridge network.

        This class is consistent with PostgresConnection in
        trove/guestagent/datastore/postgres/service.py
        """
        self.user = user
        self.password = password
        self.host = host
        self.port = port

        self.connect_str = (f"user='{self.user}' password='{self.password}' "
                            f"host='{self.host}' port='{self.port}'")

    def __enter__(self, autocommit=False):
        self.conn = psycopg2.connect(self.connect_str)
        self.conn.autocommit = autocommit
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.conn.close()

    def execute(self, statement, identifiers=None, data_values=None):
        """Execute a non-returning statement."""
        self._execute_stmt(statement, identifiers, data_values, False)

    def query(self, query, identifiers=None, data_values=None):
        """Execute a query and return the result set."""
        return self._execute_stmt(query, identifiers, data_values, True)

    def _execute_stmt(self, statement, identifiers, data_values, fetch):
        cmd = self._bind(statement, identifiers)
        with self.conn.cursor() as cursor:
            cursor.execute(cmd, data_values)
            if fetch:
                return cursor.fetchall()

    def _bind(self, statement, identifiers):
        if identifiers:
            return statement.format(*identifiers)
        return statement



# Copyright 2015 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from proboscis import SkipTest

from trove.tests.config import CONFIG
from trove.tests.scenario.helpers.test_helper import DataType
from trove.tests.scenario.runners.test_runners import TestRunner


class InstanceActionsRunner(TestRunner):

    def __init__(self):
        super(InstanceActionsRunner, self).__init__()

        self.resize_flavor_id = self._get_resize_flavor().id

    def _get_resize_flavor(self):
        if self.EPHEMERAL_SUPPORT:
            flavor_name = CONFIG.values.get(
                'instance_bigger_eph_flavor_name', 'eph.rd-smaller')
        else:
            flavor_name = CONFIG.values.get(
                'instance_bigger_flavor_name', 'm1.rd-smaller')

        return self.get_flavor(flavor_name)

    def run_add_test_data(self):
        host = self.get_instance_host(self.instance_info.id)
        self.test_helper.add_data(DataType.small, host)

    def run_verify_test_data(self):
        host = self.get_instance_host(self.instance_info.id)
        self.test_helper.verify_data(DataType.small, host)

    def run_remove_test_data(self):
        host = self.get_instance_host(self.instance_info.id)
        self.test_helper.remove_data(DataType.small, host)

    def run_instance_restart(
            self, expected_states=['REBOOT', 'HEALTHY'],
            expected_http_code=202):
        self.assert_instance_restart(self.instance_info.id, expected_states,
                                     expected_http_code)

    def assert_instance_restart(self, instance_id, expected_states,
                                expected_http_code):
        self.report.log("Testing restart on instance: %s" % instance_id)

        client = self.auth_client
        client.instances.restart(instance_id)
        self.assert_client_code(client, expected_http_code)
        self.assert_instance_action(instance_id, expected_states)

    def run_instance_resize_volume(
            self, resize_amount=1,
            expected_states=['RESIZE', 'HEALTHY'],
            expected_http_code=202):
        if self.VOLUME_SUPPORT:
            self.assert_instance_resize_volume(
                self.instance_info.id, resize_amount, expected_states,
                expected_http_code)
        else:
            raise SkipTest("Volume support is disabled.")

    def assert_instance_resize_volume(self, instance_id, resize_amount,
                                      expected_states, expected_http_code):
        self.report.log("Testing volume resize by '%d' on instance: %s"
                        % (resize_amount, instance_id))

        instance = self.get_instance(instance_id)
        old_volume_size = int(instance.volume['size'])
        new_volume_size = old_volume_size + resize_amount

        client = self.auth_client
        client.instances.resize_volume(instance_id, new_volume_size)
        self.assert_client_code(client, expected_http_code)
        self.assert_instance_action(instance_id, expected_states)

        instance = self.get_instance(instance_id)
        self.assert_equal(new_volume_size, instance.volume['size'],
                          'Unexpected new volume size')

    def run_instance_resize_flavor(self, expected_http_code=202):
        self.assert_instance_resize_flavor(
            self.instance_info.id, self.resize_flavor_id, expected_http_code)

    def assert_instance_resize_flavor(self, instance_id, resize_flavor_id,
                                      expected_http_code):
        self.report.log("Testing resize to '%s' on instance: %s" %
                        (resize_flavor_id, instance_id))
        client = self.auth_client
        client.instances.resize_instance(instance_id, resize_flavor_id)
        self.assert_client_code(client, expected_http_code)

    def run_wait_for_instance_resize_flavor(
            self, expected_states=['RESIZE', 'HEALTHY']):
        self.report.log("Waiting for resize to '%s' on instance: %s" %
                        (self.resize_flavor_id, self.instance_info.id))
        self._assert_instance_states(self.instance_info.id, expected_states)
        instance = self.get_instance(self.instance_info.id)
        self.assert_equal(self.resize_flavor_id, instance.flavor['id'],
                          'Unexpected resize flavor_id')



# Copyright 2015 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from enum import Enum
import inspect
from proboscis import SkipTest
from time import sleep


class DataType(Enum):

    """
    Represent the type of data to add to a datastore.  This allows for
    multiple 'states' of data that can be verified after actions are
    performed by Trove.
    If new entries are added here, sane values should be added to the
    _fn_data dictionary defined in TestHelper.
    """

    # micro amount of data, useful for testing datastore logging, etc.
    micro = 1
    # another micro dataset (also for datastore logging)
    micro2 = 2
    # another micro dataset (also for datastore logging)
    micro3 = 3
    # another micro dataset (also for datastore logging)
    micro4 = 4
    # very tiny amount of data, useful for testing replication
    # propagation, etc.
    tiny = 5
    # another tiny dataset (also for replication propagation)
    tiny2 = 6
    # a third tiny dataset (also for replication propagation)
    tiny3 = 7
    # a forth tiny dataset (for cluster propagation)
    tiny4 = 8
    # small amount of data (this can be added to each instance
    # after creation, for example).
    small = 9
    # large data, enough to make creating a backup take 20s or more.
    large = 10


class TestHelper(object):

    """
    Base class for all 'Helper' classes.

    The Helper classes are designed to do datastore specific work
    that can be used by multiple runner classes.  Things like adding
    data to datastores and verifying data or internal database states,
    etc. should be handled by these classes.
    """

    # Define the actions that can be done on each DataType.  When adding
    # a new action, remember to modify _data_fns
    FN_ADD = 'add'
    FN_REMOVE = 'remove'
    FN_VERIFY = 'verify'
    FN_TYPES = [FN_ADD, FN_REMOVE, FN_VERIFY]

    # Artificial 'DataType' name to use for the methods that do the
    # actual data manipulation work.
    DT_ACTUAL = 'actual'

    def __init__(self, expected_override_name, report):
        """Initialize the helper class by creating a number of stub
        functions that each datastore specific class can chose to
        override.  Basically, the functions are of the form:
            {FN_TYPE}_{DataType.name}_data
        For example:
            add_tiny_data
            add_small_data
            remove_small_data
            verify_large_data
        and so on.  Add and remove actions throw a SkipTest if not
        implemented, and verify actions by default do nothing.
        These methods, by default, call the corresponding *_actual_data()
        passing in 'data_label', 'data_start' and 'data_size' as defined
        for each DataType in the dictionary below.
        """
        super(TestHelper, self).__init__()

        self._expected_override_name = expected_override_name
        self.report = report

        # For building data access functions
        # name/fn pairs for each action
        self._data_fns = {self.FN_ADD: {},
                          self.FN_REMOVE: {},
                          self.FN_VERIFY: {}}
        # Pattern used to create the data functions.  The first parameter
        # is the function type (FN_TYPE), the second is the DataType
        # or DT_ACTUAL.
        self.data_fn_pattern = '%s_%s_data'
        # Values to distinguish between the different DataTypes.  If these
        # values don't work for a datastore, it will need to override
        # the auto-generated {FN_TYPE}_{DataType.name}_data method.
        self.DATA_START = 'start'
        self.DATA_SIZE = 'size'
        self._fn_data = {
            DataType.micro.name: {
                self.DATA_START: 100,
                self.DATA_SIZE: 10},
            DataType.micro2.name: {
                self.DATA_START: 200,
                self.DATA_SIZE: 10},
            DataType.micro3.name: {
                self.DATA_START: 300,
                self.DATA_SIZE: 10},
            DataType.micro4.name: {
                self.DATA_START: 400,
                self.DATA_SIZE: 10},
            DataType.tiny.name: {
                self.DATA_START: 1000,
                self.DATA_SIZE: 100},
            DataType.tiny2.name: {
                self.DATA_START: 2000,
                self.DATA_SIZE: 100},
            DataType.tiny3.name: {
                self.DATA_START: 3000,
                self.DATA_SIZE: 100},
            DataType.tiny4.name: {
                self.DATA_START: 4000,
                self.DATA_SIZE: 100},
            DataType.small.name: {
                self.DATA_START: 10000,
                self.DATA_SIZE: 1000},
            DataType.large.name: {
                self.DATA_START: 100000,
                self.DATA_SIZE: 100000},
        }

        self._build_data_fns()

    #################
    # Utility methods
    #################
    def get_class_name(self):
        """Builds a string of the expected class name, plus the actual one
        being used if it's not the same.
        """
        class_name_str = "'%s'" % self._expected_override_name
        if self._expected_override_name != self.__class__.__name__:
            class_name_str += ' (using %s)' % self.__class__.__name__
        return class_name_str

    ################
    # Client related
    ################
    def get_client(self, host, *args, **kwargs):
        """Gets the datastore client. This isn't cached as the
        database may be restarted in between calls, causing
        lost connection errors.
        """
        return self.create_client(host, *args, **kwargs)

    def create_client(self, host, *args, **kwargs):
        """Create a datastore client.  This is datastore specific, so this
        method should be overridden if datastore access is desired.
        """
        raise SkipTest('No client defined')

    def get_helper_credentials(self):
        """Return the credentials that the client will be using to
        access the database.
        """
        return {'name': None, 'password': None, 'database': None}

    def ping(self, host, *args, **kwargs):
        """Try to connect to a given host and perform a simple read-only
        action.

        Return True on success or False otherwise.
        """
        pass

    ##############
    # Root related
    ##############
    def get_helper_credentials_root(self):
        """Return the credentials that the client will be using to
        access the database as root.
        """
        return {'name': None, 'password': None}

    ##############
    # Data related
    ##############
    def add_data(self, data_type, host, *args, **kwargs):
        """Adds data of type 'data_type' to the database.  Descendant
        classes should implement a function 'add_actual_data' that has the
        following signature:
            def add_actual_data(
                self,        # standard self reference
                data_label,  # label used to identify the 'type' to add
                data_start,  # a start count
                data_size,   # a size to use
                host,        # the host to add the data to
                *args,       # for possible future expansion
                **kwargs     # for possible future expansion
            ):
        The data_label could be used to create a database or a table if the
        datastore supports that.  The data_start and data_size values are
        designed not to overlap, such that all the data could be stored
        in a single namespace (for example, creating ids from data_start
        to data_start + data_size).

        Since this method may be called multiple times, the
        'add_actual_data' function should be idempotent.
        """
        self._perform_data_action(self.FN_ADD, data_type.name,
                                  host, *args, **kwargs)

    def remove_data(self, data_type, host, *args, **kwargs):
        """Removes all data associated with 'data_type'.  See
        instructions for 'add_data' for implementation guidance.
        """
        self._perform_data_action(self.FN_REMOVE, data_type.name,
                                  host, *args, **kwargs)

    def verify_data(self, data_type, host, *args, **kwargs):
        """Verify that the data of type 'data_type' exists in the
        datastore.  This can be done by testing edge cases, and possibly
        some random elements within the set.  See
        instructions for 'add_data' for implementation guidance.
        """
        self._perform_data_action(self.FN_VERIFY, data_type.name,
                                  host, *args, **kwargs)

    def _perform_data_action(self, fn_type, fn_name, host,
                             *args, **kwargs):
        """By default, the action is attempted 10 times, sleeping for 3
        seconds between each attempt.  This can be controlled by the
        retry_count and retry_sleep kwarg values.
        """
        retry_count = kwargs.pop('retry_count', 10) or 0
        retry_sleep = kwargs.pop('retry_sleep', 3) or 0

        fns = self._data_fns[fn_type]
        data_fn_name = self.data_fn_pattern % (fn_type, fn_name)
        attempts = -1
        while True:
            attempts += 1
            try:
                fns[data_fn_name](self, host, *args, **kwargs)
                break
            except SkipTest:
                raise
            except Exception as ex:
                self.report.log("Attempt %d to %s data type %s failed\n%s"
                                % (attempts, fn_type, fn_name, ex))
                if attempts > retry_count:
                    raise RuntimeError("Error calling %s from class %s - %s" %
                                       (data_fn_name, self.__class__.__name__,
                                        ex))
                self.report.log("Trying again (after %d second sleep)" %
                                retry_sleep)
                sleep(retry_sleep)

    def _build_data_fns(self):
        """Build the base data functions specified by FN_TYPE_*
        for each of the types defined in the DataType class.  For example,
        'add_small_data' and 'verify_large_data'.  These
        functions are set to call '*_actual_data' and will pass in
        sane values for label, start and size.  The '*_actual_data'
        methods should be overwritten by a descendant class, and are the
        ones that do the actual work.
        The original 'add_small_data', etc. methods can also be overridden
        if needed, and those overwritten functions will be bound before
        calling any data functions such as 'add_data' or 'remove_data'.
        """
        for fn_type in self.FN_TYPES:
            fn_dict = self._data_fns[fn_type]
            for data_type in DataType:
                self._data_fn_builder(fn_type, data_type.name, fn_dict)
            self._data_fn_builder(fn_type, self.DT_ACTUAL, fn_dict)
        self._override_data_fns()

    def _data_fn_builder(self, fn_type, fn_name, fn_dict):
        """Builds the actual function with a SkipTest exception,
        and changes the name to reflect the pattern.
        """
        data_fn_name = self.data_fn_pattern % (fn_type, fn_name)

        # Build the overridable 'actual' Data Manipulation methods
        if fn_name == self.DT_ACTUAL:
            def data_fn(self, data_label, data_start, data_size, host,
                        *args, **kwargs):
                # default action is to skip the test
                cls_str = ''
                if self._expected_override_name != self.__class__.__name__:
                    cls_str = (' (%s not loaded)' %
                               self._expected_override_name)
                raise SkipTest("Data function '%s' not found in '%s'%s" % (
                    data_fn_name, self.__class__.__name__, cls_str))
        else:
            def data_fn(self, host, *args, **kwargs):
                # call the corresponding 'actual' method
                fns = self._data_fns[fn_type]
                var_dict = self._fn_data[fn_name]
                data_start = var_dict[self.DATA_START]
                data_size = var_dict[self.DATA_SIZE]
                actual_fn_name = self.data_fn_pattern % (
                    fn_type, self.DT_ACTUAL)
                try:
                    fns[actual_fn_name](self, fn_name, data_start, data_size,
                                        host, *args, **kwargs)
                except SkipTest:
                    raise
                except Exception as ex:
                    raise RuntimeError("Error calling %s from class %s: %s" % (
                        data_fn_name, self.__class__.__name__, ex))

        data_fn.__name__ = data_fn.func_name = data_fn_name
        fn_dict[data_fn_name] = data_fn

    def _override_data_fns(self):
        """Bind the override methods to the dict."""
        members = inspect.getmembers(self.__class__,
                                     predicate=inspect.ismethod)
        for fn_type in self.FN_TYPES:
            fns = self._data_fns[fn_type]
            for name, fn in members:
                if name in fns:
                    fns[name] = fn

    #######################
    # Database/User related
    #######################
    def get_valid_database_definitions(self):
        """Return a list of valid database JSON definitions.
        These definitions will be used by tests that create databases.
        Return an empty list if the datastore does not support databases.
        """
        return list()

    def get_valid_user_definitions(self):
        """Return a list of valid user JSON definitions.
         These definitions will be used by tests that create users.
         Return an empty list if the datastore does not support users.
         """
        return list()

    def get_non_existing_database_definition(self):
        """Return a valid JSON definition for a non-existing database.
         This definition will be used by negative database tests.
         The database will not be created by any of the tests.
         Return None if the datastore does not support databases.
         """
        valid_defs = self.get_valid_database_definitions()
        return self._get_non_existing_definition(valid_defs)

    def get_non_existing_user_definition(self):
        """Return a valid JSON definition for a non-existing user.
         This definition will be used by negative user tests.
         The user will not be created by any of the tests.
         Return None if the datastore does not support users.
         """
        valid_defs = self.get_valid_user_definitions()
        return self._get_non_existing_definition(valid_defs)

    def _get_non_existing_definition(self, existing_defs):
        """This will create a unique definition for a non-existing object
        by randomizing one of an existing object.
        """
        if existing_defs:
            non_existing_def = dict(existing_defs[0])
            while non_existing_def in existing_defs:
                non_existing_def = self._randomize_on_name(non_existing_def)
            return non_existing_def

        return None

    def _randomize_on_name(self, definition):
        def_copy = dict(definition)
        def_copy['name'] = ''.join([def_copy['name'], 'rnd'])
        return def_copy

    #############################
    # Configuration Group related
    #############################
    def get_dynamic_group(self):
        """Return a definition of a dynamic configuration group.
        A dynamic group should contain only properties that do not require
        database restart.
        Return an empty dict if the datastore does not have any.
        """
        return dict()

    def get_non_dynamic_group(self):
        """Return a definition of a non-dynamic configuration group.
        A non-dynamic group has to include at least one property that requires
        database restart.
        Return an empty dict if the datastore does not have any.
        """
        return dict()

    def get_invalid_groups(self):
        """Return a list of configuration groups with invalid values.
        An empty list indicates that no 'invalid' tests should be run.
        """
        return []

    def get_configuration_value(self, property_name, host, *args, **kwargs):
        """Use the client to retrieve the value of a given configuration
        property.
        """
        raise SkipTest("Runtime configuration retrieval not implemented in %s"
                       % self.get_class_name())

    ###################
    # Guest Log related
    ###################
    def get_exposed_log_list(self):
        """Return the list of exposed logs for the datastore.  This
        method shouldn't need to be overridden.
        """
        logs = []
        try:
            logs.extend(self.get_exposed_user_log_names())
        except SkipTest:
            pass
        try:
            logs.extend(self.get_exposed_sys_log_names())
        except SkipTest:
            pass

        return logs

    def get_full_log_list(self):
        """Return the full list of all logs for the datastore.  This
        method shouldn't need to be overridden.
        """
        logs = self.get_exposed_log_list()
        try:
            logs.extend(self.get_unexposed_user_log_names())
        except SkipTest:
            pass
        try:
            logs.extend(self.get_unexposed_sys_log_names())
        except SkipTest:
            pass

        return logs

    # Override these guest log methods if needed
    def get_exposed_user_log_names(self):
        """Return the names of the user logs that are visible to all users.
        The first log name will be used for tests.
        """
        raise SkipTest("No exposed user log names defined.")

    def get_unexposed_user_log_names(self):
        """Return the names of the user logs that not visible to all users.
        The first log name will be used for tests.
        """
        raise SkipTest("No unexposed user log names defined.")

    def get_exposed_sys_log_names(self):
        """Return the names of SYS logs that are visible to all users.
        The first log name will be used for tests.
        """
        raise SkipTest("No exposed sys log names defined.")

    def get_unexposed_sys_log_names(self):
        """Return the names of the sys logs that not visible to all users.
        The first log name will be used for tests.
        """
        return ['guest']

    def log_enable_requires_restart(self):
        """Returns whether enabling or disabling a USER log requires a
        restart of the datastore.
        """
        return False

    ################
    # Module related
    ################
    def get_valid_module_type(self):
        """Return a valid module type."""
        return "Ping"

    #################
    # Cluster related
    #################
    def get_cluster_types(self):
        """Returns a list of cluster type lists to use when creating instances.
        The list should be the same size as the number of cluster instances
        that will be created.  If not specified, no types are sent to
        cluster-create.  Cluster grow uses the first type in the list for the
        first instance, and doesn't use anything for the second instance
        (i.e. doesn't pass in anything for 'type').
        An example for this method would be:
            return [['data', 'other_type'], ['third_type']]
        """
        return None



# Copyright 2013 OpenStack Foundation
# Copyright 2013 Rackspace Hosting
# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import tempfile

from oslo_log import log as logging

from trove.common import cfg
from trove.common import configurations
from trove.common import exception
from trove.common import utils
from trove.common.notification import EndNotification
from trove.guestagent import guest_log
from trove.guestagent.common import operating_system
from trove.guestagent.datastore import manager
from trove.guestagent.utils import docker as docker_util
from trove.guestagent.utils import mysql as mysql_util
from trove.instance import service_status

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class MySqlManager(manager.Manager):
    def __init__(self, mysql_app, mysql_app_status, mysql_admin,
                 manager_name='mysql'):
        super(MySqlManager, self).__init__(manager_name)

        self.app = mysql_app
        self.status = mysql_app_status
        self.adm = mysql_admin
        self.volume_do_not_start_on_reboot = False

    @property
    def configuration_manager(self):
        return self.app.configuration_manager

    def get_service_status(self):
        try:
            with mysql_util.SqlClient(self.app.get_engine()) as client:
                cmd = "SELECT 1;"
                client.execute(cmd)

            LOG.debug("Database service check: database query is responsive")
            return service_status.ServiceStatuses.HEALTHY
        except Exception:
            return super(MySqlManager, self).get_service_status()

    def get_start_db_params(self, data_dir):
        return f'--datadir={data_dir}'

    def do_prepare(self, context, packages, databases, memory_mb, users,
                   device_path, mount_point, backup_info,
                   config_contents, root_password, overrides,
                   cluster_config, snapshot, ds_version=None):
        """This is called from prepare in the base class."""
        data_dir = mount_point + '/data'
        self.app.stop_db()
        operating_system.ensure_directory(data_dir,
                                          user=CONF.database_service_uid,
                                          group=CONF.database_service_uid,
                                          as_root=True)
        # This makes sure the include dir is created.
        self.app.set_data_dir(data_dir)

        # Prepare mysql configuration
        LOG.info('Preparing database configuration')
        self.app.configuration_manager.reset_configuration(config_contents)
        self.app.update_overrides(overrides)

        # Restore data from backup and reset root password
        if backup_info:
            self.perform_restore(context, data_dir, backup_info)
            self.reset_password_for_restore(ds_version=ds_version,
                                            data_dir=data_dir)

        # Start database service.
        command = self.get_start_db_params(data_dir)
        self.app.start_db(ds_version=ds_version, command=command)

        self.app.secure()
        enable_remote_root = (backup_info and self.adm.is_root_enabled())
        if enable_remote_root:
            self.status.report_root(context)
        else:
            self.app.secure_root()

        if snapshot:
            # This instance is a replication slave
            self.attach_replica(context, snapshot, snapshot['config'])

    def create_backup(self, context, backup_info):
        """Create backup for the database.

        :param context: User context object.
        :param backup_info: a dictionary containing the db instance id of the
                            backup task, location, type, and other data.
        """
        LOG.info(f"Creating backup {backup_info['id']}")
        with EndNotification(context):
            # Set /var/run/mysqld to allow localhost access.
            volumes_mapping = {
                '/var/lib/mysql': {'bind': '/var/lib/mysql', 'mode': 'rw'},
                "/var/run/mysqld": {"bind": "/var/run/mysqld", "mode": "ro"},
                '/tmp': {'bind': '/tmp', 'mode': 'rw'}
            }
            self.app.create_backup(context, backup_info,
                                   volumes_mapping=volumes_mapping,
                                   need_dbuser=True)

    def get_datastore_log_defs(self):
        owner = cfg.get_configuration_property('database_service_uid')
        datastore_dir = self.app.get_data_dir()
        server_section = configurations.MySQLConfParser.SERVER_CONF_SECTION
        long_query_time = CONF.get(self.manager).get(
            'guest_log_long_query_time') / 1000
        general_log_file = self.build_log_file_name(
            self.GUEST_LOG_DEFS_GENERAL_LABEL, owner,
            datastore_dir=datastore_dir)
        error_log_file = self.validate_log_file('/var/log/mysqld.log', owner)
        slow_query_log_file = self.build_log_file_name(
            self.GUEST_LOG_DEFS_SLOW_QUERY_LABEL, owner,
            datastore_dir=datastore_dir)
        return {
            self.GUEST_LOG_DEFS_GENERAL_LABEL: {
                self.GUEST_LOG_TYPE_LABEL: guest_log.LogType.USER,
                self.GUEST_LOG_USER_LABEL: owner,
                self.GUEST_LOG_FILE_LABEL: general_log_file,
                self.GUEST_LOG_SECTION_LABEL: server_section,
                self.GUEST_LOG_ENABLE_LABEL: {
                    'general_log': 'on',
                    'general_log_file': general_log_file,
                    'log_output': 'file',
                },
                self.GUEST_LOG_DISABLE_LABEL: {
                    'general_log': 'off',
                },
            },
            self.GUEST_LOG_DEFS_SLOW_QUERY_LABEL: {
                self.GUEST_LOG_TYPE_LABEL: guest_log.LogType.USER,
                self.GUEST_LOG_USER_LABEL: owner,
                self.GUEST_LOG_FILE_LABEL: slow_query_log_file,
                self.GUEST_LOG_SECTION_LABEL: server_section,
                self.GUEST_LOG_ENABLE_LABEL: {
                    'slow_query_log': 'on',
                    'slow_query_log_file': slow_query_log_file,
                    'long_query_time': long_query_time,
                },
                self.GUEST_LOG_DISABLE_LABEL: {
                    'slow_query_log': 'off',
                },
            },
            self.GUEST_LOG_DEFS_ERROR_LABEL: {
                self.GUEST_LOG_TYPE_LABEL: guest_log.LogType.SYS,
                self.GUEST_LOG_USER_LABEL: owner,
                self.GUEST_LOG_FILE_LABEL: error_log_file,
            },
        }

    def is_log_enabled(self, logname):
        if logname == self.GUEST_LOG_DEFS_GENERAL_LABEL:
            value = self.configuration_manager.get_value(
                'general_log', section='mysqld', default='off')
            LOG.debug(f"The config value of general_log is {value}")
            return value == 'on'
        elif logname == self.GUEST_LOG_DEFS_SLOW_QUERY_LABEL:
            value = self.configuration_manager.get_value(
                'slow_query_log', section='mysqld', default='off')
            LOG.debug(f"The config value of slow_query_log is {value}")
            return value == 'on'

        return False

    def apply_overrides(self, context, overrides):
        LOG.info("Applying database config.")
        self.app.apply_overrides(overrides)
        LOG.info("Finished applying database config.")

    def reset_password_for_restore(self, ds_version=None,
                                   data_dir='/var/lib/mysql/data'):
        """Reset the root password after restore the db data.

        We create a temporary database container by running mysqld_safe to
        reset the root password.
        """
        LOG.info('Starting to reset password for restore')

        try:
            root_pass = self.app.get_auth_password(file="root.cnf")
        except exception.UnprocessableEntity:
            root_pass = utils.generate_random_password()
            self.app.save_password('root', root_pass)

        with tempfile.NamedTemporaryFile(mode='w') as init_file, \
            tempfile.NamedTemporaryFile(suffix='.err') as err_file:
            operating_system.write_file(
                init_file.name,
                f"ALTER USER 'root'@'localhost' IDENTIFIED BY '{root_pass}';"
            )
            command = (
                f'mysqld_safe --init-file={init_file.name} '
                f'--log-error={err_file.name} '
                f'--datadir={data_dir}'
            )
            extra_volumes = {
                init_file.name: {"bind": init_file.name, "mode": "rw"},
                err_file.name: {"bind": err_file.name, "mode": "rw"},
            }

            # Allow database service user to access the temporary files.
            for file in [init_file.name, err_file.name]:
                operating_system.chmod(file,
                                       operating_system.FileMode.SET_ALL_RWX(),
                                       force=True, as_root=True)

            try:
                self.app.start_db(ds_version=ds_version, command=command,
                                  extra_volumes=extra_volumes)
            except Exception as err:
                LOG.error('Failed to reset password for restore, error: %s',
                          str(err))
                LOG.debug('Content in init error log file: %s',
                          err_file.read())
                raise err
            finally:
                LOG.debug(
                    'The init container log: %s',
                    docker_util.get_container_logs(self.app.docker_client)
                )
                docker_util.remove_container(self.app.docker_client)

        LOG.info('Finished to reset password for restore')

    def _validate_slave_for_replication(self, context, replica_info):
        if replica_info['replication_strategy'] != self.replication_strategy:
            raise exception.IncompatibleReplicationStrategy(
                replica_info.update({
                    'guest_strategy': self.replication_strategy
                }))

        volume_stats = self.get_filesystem_stats(context, None)
        if (volume_stats.get('total', 0.0) <
            replica_info['dataset']['dataset_size']):
            raise exception.InsufficientSpaceForReplica(
                replica_info.update({
                    'slave_volume_size': volume_stats.get('total', 0.0)
                }))

    def attach_replica(self, context, replica_info, slave_config, **kwargs):
        LOG.info("Attaching replica, replica_info: %s", replica_info)
        try:
            if 'replication_strategy' in replica_info:
                self._validate_slave_for_replication(context, replica_info)

            self.replication.enable_as_slave(self.app, replica_info,
                                             slave_config)
        except Exception as err:
            LOG.error("Error enabling replication, error: %s", str(err))
            self.status.set_status(service_status.ServiceStatuses.FAILED)
            raise

    def make_read_only(self, context, read_only):
        LOG.info("Executing make_read_only(%s)", read_only)
        self.app.make_read_only(read_only)

    def get_latest_txn_id(self, context):
        LOG.info("Calling get_latest_txn_id.")
        return self.app.get_latest_txn_id()

    def get_last_txn(self, context):
        LOG.info("Calling get_last_txn")
        return self.app.get_last_txn()

    def wait_for_txn(self, context, txn):
        LOG.info("Calling wait_for_txn.")
        self.app.wait_for_txn(txn)

    def upgrade(self, context, upgrade_info):
        """Upgrade the database."""
        LOG.info('Starting to upgrade database, upgrade_info: %s',
                 upgrade_info)
        self.app.upgrade(upgrade_info)

    def rebuild(self, context, ds_version, config_contents=None,
                config_overrides=None):
        """Restore datastore service after instance rebuild."""
        LOG.info("Starting to restore database service")
        self.status.begin_install()

        mount_point = CONF.get(CONF.datastore_manager).mount_point
        data_dir = mount_point + '/data'
        operating_system.ensure_directory(data_dir,
                                          user=CONF.database_service_uid,
                                          group=CONF.database_service_uid,
                                          as_root=True)
        # This makes sure the include dir is created.
        self.app.set_data_dir(data_dir)

        try:
            # Prepare mysql configuration
            LOG.debug('Preparing database configuration')
            self.app.configuration_manager.reset_configuration(config_contents)
            self.app.update_overrides(config_overrides)

            # Start database service.
            command = self.get_start_db_params(data_dir)
            self.app.start_db(ds_version=ds_version, command=command)
        except Exception as e:
            LOG.error(f"Failed to restore database service after rebuild, "
                      f"error: {str(e)}")
            self.prepare_error = True
            raise
        finally:
            self.status.end_install(error_occurred=self.prepare_error)



# Copyright 2016 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from sqlalchemy.schema import Column
from sqlalchemy.schema import MetaData
from sqlalchemy.sql.expression import update
from sqlalchemy import text

from trove.db.sqlalchemy.migrate_repo.schema import Boolean
from trove.db.sqlalchemy.migrate_repo.schema import Integer
from trove.db.sqlalchemy.migrate_repo.schema import Table
from trove.db.sqlalchemy.migrate_repo.schema import Text


COLUMN_NAME_1 = 'priority_apply'
COLUMN_NAME_2 = 'apply_order'
COLUMN_NAME_3 = 'is_admin'


def upgrade(migrate_engine):
    meta = MetaData()
    meta.bind = migrate_engine
    modules = Table('modules', meta, autoload=True)
    is_nullable = True if migrate_engine.name == "sqlite" else False
    column = Column(COLUMN_NAME_1, Boolean(), nullable=is_nullable, default=0)
    modules.create_column(column)
    column = Column(COLUMN_NAME_2, Integer(), nullable=is_nullable, default=5)
    modules.create_column(column)
    column = Column(COLUMN_NAME_3, Boolean(), nullable=is_nullable, default=0)
    modules.create_column(column)
    modules.c.contents.alter(Text(length=4294967295))
    # mark all non-visible, auto-apply and all-tenant modules as is_admin
    update(table=modules,
           values=dict(is_admin=1),
           whereclause=text("visible=0 or auto_apply=1 or tenant_id is null")
           ).execute()



# Copyright 2013 OpenStack Foundation
# Copyright 2013 Rackspace Hosting
# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import functools
import gettext
import os
import sys
import traceback

import eventlet
from oslo_log import log as logging
import proboscis
import urllib
import wsgi_intercept
from wsgi_intercept.httplib2_intercept import install as wsgi_install

from trove.common import cfg
from trove.common.rpc import service as rpc_service
from trove.common.rpc import version as rpc_version
from trove.common import utils
from trove import rpc
from trove.tests.config import CONFIG
from trove.tests import root_logger

eventlet.monkey_patch(thread=False)

CONF = cfg.CONF
original_excepthook = sys.excepthook


def add_support_for_localization():
    """Adds support for localization in the logging.

    If ../nova/__init__.py exists, add ../ to Python search path, so that
    it will override what happens to be installed in
    /usr/(local/)lib/python...

    """
    path = os.path.join(os.path.abspath(sys.argv[0]), os.pardir, os.pardir)
    possible_topdir = os.path.normpath(path)
    if os.path.exists(os.path.join(possible_topdir, 'nova', '__init__.py')):
        sys.path.insert(0, possible_topdir)

    gettext.install('nova')


def initialize_trove(config_file):
    from trove.common import pastedeploy

    root_logger.DefaultRootLogger()

    cfg.CONF(args=[],
             project='trove',
             default_config_files=[config_file])
    logging.setup(CONF, None)
    topic = CONF.taskmanager_queue
    rpc.init(CONF)

    taskman_service = rpc_service.RpcService(
        CONF.taskmanager_rpc_encr_key, topic=topic,
        rpc_api_version=rpc_version.RPC_API_VERSION,
        manager='trove.taskmanager.manager.Manager')
    taskman_service.start()

    return pastedeploy.paste_deploy_app(config_file, 'trove', {})


def datastore_init():
    # Adds the datastore for mysql (needed to make most calls work).
    from trove.configuration.models import DatastoreConfigurationParameters
    from trove.datastore import models

    models.DBDatastore.create(
        id=CONFIG.dbaas_datastore_id, name=CONFIG.dbaas_datastore,
        default_version_id=CONFIG.dbaas_datastore_version_id)

    models.DBDatastore.create(id=utils.generate_uuid(),
                              name=CONFIG.dbaas_datastore_name_no_versions,
                              default_version_id=None)

    main_dsv = models.DBDatastoreVersion.create(
        id=CONFIG.dbaas_datastore_version_id,
        datastore_id=CONFIG.dbaas_datastore_id,
        name=CONFIG.dbaas_datastore_version,
        manager="mysql",
        image_id='c00000c0-00c0-0c00-00c0-000c000000cc',
        packages='test packages',
        active=1)
    models.DBDatastoreVersion.create(
        id="d00000d0-00d0-0d00-00d0-000d000000dd",
        datastore_id=CONFIG.dbaas_datastore_id,
        name='mysql_inactive_version', manager="mysql",
        image_id='c00000c0-00c0-0c00-00c0-000c000000cc',
        packages=None, active=0)

    def add_parm(name, data_type, max_size, min_size=0, restart_required=0):
        DatastoreConfigurationParameters.create(
            datastore_version_id=main_dsv.id,
            name=name,
            restart_required=restart_required,
            max_size=max_size,
            min_size=0,
            data_type=data_type,
            deleted=0,
            deleted_at=None)

    add_parm('key_buffer_size', 'integer', 4294967296)
    add_parm('connect_timeout', 'integer', 65535)
    add_parm('join_buffer_size', 'integer', 4294967296)
    add_parm('local_infile', 'integer', 1)
    add_parm('collation_server', 'string', None, None)
    add_parm('innodb_buffer_pool_size', 'integer', 57671680,
             restart_required=1)


def initialize_database():
    from trove.db import get_db_api
    from trove.db.sqlalchemy import session
    db_api = get_db_api()
    db_api.drop_db(CONF)  # Destroys the database, if it exists.
    db_api.db_sync(CONF)
    session.configure_db(CONF)
    datastore_init()
    db_api.configure_db(CONF)


def initialize_fakes(app):
    # Set up WSGI interceptor. This sets up a fake host that responds each
    # time httplib tries to communicate to localhost, port 8779.
    def wsgi_interceptor(*args, **kwargs):

        def call_back(env, start_response):
            path_info = env.get('PATH_INFO')
            if path_info:
                env['PATH_INFO'] = urllib.parse.unquote(path_info)
            return app.__call__(env, start_response)

        return call_back

    wsgi_intercept.add_wsgi_intercept('localhost',
                                      CONF.bind_port,
                                      wsgi_interceptor)
    from trove.tests.util import event_simulator
    event_simulator.monkey_patch()
    from trove.tests.fakes import taskmanager
    taskmanager.monkey_patch()


def parse_args_for_test_config():
    test_conf = 'etc/tests/localhost.test.conf'
    repl = False
    new_argv = []
    for index in range(len(sys.argv)):
        arg = sys.argv[index]
        print(arg)
        if arg[:14] == "--test-config=":
            test_conf = arg[14:]
        elif arg == "--repl":
            repl = True
        else:
            new_argv.append(arg)
    sys.argv = new_argv
    return test_conf, repl


def run_tests(repl):
    """Runs all of the tests."""

    if repl:
        # Actually show errors in the repl.
        sys.excepthook = original_excepthook

        def no_thanks(exit_code):
            print("Tests finished with exit code %d." % exit_code)
        sys.exit = no_thanks

    proboscis.TestProgram().run_and_exit()

    if repl:
        import code
        code.interact()


def import_tests():
    # F401 unused imports needed for tox tests
    from trove.tests.api import backups  # noqa
    from trove.tests.api import configurations  # noqa
    from trove.tests.api import databases  # noqa
    from trove.tests.api import datastores  # noqa
    from trove.tests.api import instances as rd_instances  # noqa
    from trove.tests.api import instances_actions as rd_actions  # noqa
    from trove.tests.api import instances_delete  # noqa
    from trove.tests.api import instances_resize  # noqa
    from trove.tests.api import limits  # noqa
    from trove.tests.api.mgmt import instances_actions as mgmt_actions  # noqa
    from trove.tests.api import replication  # noqa
    from trove.tests.api import root  # noqa
    from trove.tests.api import user_access  # noqa
    from trove.tests.api import users  # noqa
    from trove.tests.api import versions  # noqa
    from trove.tests.db import migrations  # noqa


def main(import_func):
    try:
        wsgi_install()
        add_support_for_localization()
        # Load Trove app
        # Paste file needs absolute path
        config_file = os.path.realpath('etc/trove/trove.conf.test')
        # 'etc/trove/test-api-paste.ini'
        app = initialize_trove(config_file)
        # Initialize sqlite database.
        initialize_database()
        # Swap out WSGI, httplib, and other components with test doubles.
        initialize_fakes(app)

        # Initialize the test configuration.
        test_config_file, repl = parse_args_for_test_config()
        CONFIG.load_from_file(test_config_file)

        import_func()

        from trove.tests.util import event_simulator
        event_simulator.run_main(functools.partial(run_tests, repl))

    except Exception as e:
        # Printing the error manually like this is necessary due to oddities
        # with sys.excepthook.
        print("Run tests failed: %s" % e)
        traceback.print_exc()
        raise


if __name__ == "__main__":
    main(import_tests)



# Copyright 2020 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from unittest import mock

from trove.datastore import models as ds_models
from trove.extensions.mgmt.instances import service as ins_service
from trove.instance import models as ins_models
from trove.instance import service_status as srvstatus
from trove.tests.unittests import trove_testtools
from trove.tests.unittests.util import util


class TestMgmtInstanceController(trove_testtools.TestCase):
    @classmethod
    def setUpClass(cls):
        util.init_db()
        cls.controller = ins_service.MgmtInstanceController()

        cls.ds_name = cls.random_name('datastore')
        ds_models.update_datastore(name=cls.ds_name, default_version=None)
        ds_models.update_datastore_version(
            cls.ds_name, 'test_version', 'mysql', cls.random_uuid(), '', '', 1)
        cls.ds = ds_models.Datastore.load(cls.ds_name)
        cls.ds_version = ds_models.DatastoreVersion.load(cls.ds,
                                                         'test_version')

        cls.ins_name = cls.random_name('instance')
        cls.project_id = cls.random_uuid()
        cls.server_id = cls.random_uuid()
        cls.instance = ins_models.DBInstance.create(
            name=cls.ins_name, flavor_id=cls.random_uuid(),
            tenant_id=cls.project_id,
            volume_size=1,
            datastore_version_id=cls.ds_version.id,
            task_status=ins_models.InstanceTasks.BUILDING,
            compute_instance_id=cls.server_id
        )
        ins_models.InstanceServiceStatus.create(
            instance_id=cls.instance.id,
            status=srvstatus.ServiceStatuses.NEW
        )

        super(TestMgmtInstanceController, cls).setUpClass()

    @classmethod
    def tearDownClass(cls):
        util.cleanup_db()

        super(TestMgmtInstanceController, cls).tearDownClass()

    @mock.patch('trove.common.clients.create_nova_client')
    def test_index_project_id(self, mock_create_client):
        req = mock.MagicMock()
        req.GET = {
            'project_id': self.project_id
        }

        mock_nova_client = mock.MagicMock()
        mock_nova_client.servers.list.return_value = [
            mock.MagicMock(id=self.server_id)]
        mock_create_client.return_value = mock_nova_client

        result = self.controller.index(req, mock.ANY)

        self.assertEqual(200, result.status)
        data = result.data(None)
        self.assertEqual(1, len(data['instances']))

        req.GET = {
            'project_id': self.random_uuid()
        }

        result = self.controller.index(req, mock.ANY)

        self.assertEqual(200, result.status)
        data = result.data(None)
        self.assertEqual(0, len(data['instances']))
